from types import MappingProxyType
from typing import Any, Literal, Optional, TypeVar, Union, overload

from pydantic import Discriminator
from typing_extensions import Annotated

from bioimageio.spec._internal.validation_context import ValidationContext

from ._description_impl import DISCOVER, build_description_impl, get_rd_class_impl
from ._internal.common_nodes import InvalidDescr
from ._internal.io import BioimageioYamlContent, BioimageioYamlContentView
from ._internal.types import FormatVersionPlaceholder
from ._internal.validation_context import get_validation_context
from .application import (
    AnyApplicationDescr,
    ApplicationDescr,
    ApplicationDescr_v0_2,
    ApplicationDescr_v0_3,
)
from .dataset import AnyDatasetDescr, DatasetDescr, DatasetDescr_v0_2, DatasetDescr_v0_3
from .generic import AnyGenericDescr, GenericDescr, GenericDescr_v0_2, GenericDescr_v0_3
from .model import AnyModelDescr, ModelDescr, ModelDescr_v0_4, ModelDescr_v0_5
from .notebook import (
    AnyNotebookDescr,
    NotebookDescr,
    NotebookDescr_v0_2,
    NotebookDescr_v0_3,
)
from .summary import ValidationSummary

LATEST: Literal["latest"] = "latest"
"""placeholder for the latest available format version"""


LatestResourceDescr = Union[
    Annotated[
        Union[
            ApplicationDescr,
            DatasetDescr,
            ModelDescr,
            NotebookDescr,
        ],
        Discriminator("type"),
    ],
    GenericDescr,
]
"""A resource description following the latest specification format"""


SpecificResourceDescr = Annotated[
    Union[
        AnyApplicationDescr,
        AnyDatasetDescr,
        AnyModelDescr,
        AnyNotebookDescr,
    ],
    Discriminator("type"),
]
"""Any of the implemented, non-generic resource descriptions"""

ResourceDescr = Union[SpecificResourceDescr, AnyGenericDescr]
"""Any of the implemented resource descriptions"""


def dump_description(
    rd: Union[ResourceDescr, InvalidDescr],
    /,
    *,
    exclude_unset: bool = True,
    exclude_defaults: bool = False,
) -> BioimageioYamlContent:
    """Converts a resource to a dictionary containing only simple types that can directly be serialzed to YAML.

    Args:
        rd: bioimageio resource description
        exclude_unset: Exclude fields that have not explicitly be set.
        exclude_defaults: Exclude fields that have the default value (even if set explicitly).
    """
    return rd.model_dump(
        mode="json", exclude_unset=exclude_unset, exclude_defaults=exclude_defaults
    )


RD = TypeVar("RD", bound=ResourceDescr)


LATEST_DESCRIPTIONS_MAP = MappingProxyType(
    {
        None: GenericDescr,
        "generic": GenericDescr,
        "application": ApplicationDescr,
        "dataset": DatasetDescr,
        "notebook": NotebookDescr,
        "model": ModelDescr,
    }
)
DESCRIPTIONS_MAP = MappingProxyType(
    {
        None: MappingProxyType(
            {
                "0.2": GenericDescr_v0_2,
                "0.3": GenericDescr_v0_3,
                "latest": GenericDescr,
            }
        ),
        "generic": MappingProxyType(
            {
                "0.2": GenericDescr_v0_2,
                "0.3": GenericDescr_v0_3,
                "latest": GenericDescr,
            }
        ),
        "application": MappingProxyType(
            {
                "0.2": ApplicationDescr_v0_2,
                "0.3": ApplicationDescr_v0_3,
                "latest": ApplicationDescr,
            }
        ),
        "dataset": MappingProxyType(
            {
                "0.2": DatasetDescr_v0_2,
                "0.3": DatasetDescr_v0_3,
                "latest": DatasetDescr,
            }
        ),
        "notebook": MappingProxyType(
            {
                "0.2": NotebookDescr_v0_2,
                "0.3": NotebookDescr_v0_3,
                "latest": NotebookDescr,
            }
        ),
        "model": MappingProxyType(
            {
                "0.3": ModelDescr_v0_4,
                "0.4": ModelDescr_v0_4,
                "0.5": ModelDescr_v0_5,
                "latest": ModelDescr,
            }
        ),
    }
)
"""A mapping to determine the appropriate Description class
 for a given **type** and **format_version**."""


def _get_rd_class(typ: Any, format_version: Any, fallback_to_latest: bool):
    return get_rd_class_impl(
        typ, format_version, DESCRIPTIONS_MAP, fallback_to_latest=fallback_to_latest
    )


@overload
def build_description(
    content: BioimageioYamlContentView,
    /,
    *,
    context: Optional[ValidationContext] = None,
    format_version: Literal["latest"],
) -> Union[LatestResourceDescr, InvalidDescr]: ...


@overload
def build_description(
    content: BioimageioYamlContentView,
    /,
    *,
    context: Optional[ValidationContext] = None,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
) -> Union[ResourceDescr, InvalidDescr]: ...


def build_description(
    content: BioimageioYamlContentView,
    /,
    *,
    context: Optional[ValidationContext] = None,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
) -> Union[ResourceDescr, InvalidDescr]:
    """build a bioimage.io resource description from an RDF's content.

    Use `load_description` if you want to build a resource description from an rdf.yaml
    or bioimage.io zip-package.

    Args:
        content: loaded rdf.yaml file (loaded with YAML, not bioimageio.spec)
        context: validation context to use during validation
        format_version:
            (optional) use this argument to load the resource and
            convert its metadata to a higher format_version.
            Note:
            - Use "latest" to convert to the latest available format version.
            - Use "discover" to use the format version specified in the RDF.
            - Only considers major.minor format version, ignores patch version.
            - Conversion to lower format versions is not supported.

    Returns:
        An object holding all metadata of the bioimage.io resource

    """

    return build_description_impl(
        content,
        context=context,
        format_version=format_version,
        get_rd_class=_get_rd_class,
    )


def validate_format(
    data: BioimageioYamlContent,
    /,
    *,
    format_version: Union[Literal["discover", "latest"], str] = DISCOVER,
    context: Optional[ValidationContext] = None,
) -> ValidationSummary:
    """Validate a dictionary holding a bioimageio description.
    See `bioimagieo.spec.load_description_and_validate_format_only`
    to validate a file source.

    Args:
        data: Dictionary holding the raw bioimageio.yaml content.
        format_version:
            Format version to (update to and) use for validation.
            Note:
            - Use "latest" to convert to the latest available format version.
            - Use "discover" to use the format version specified in the RDF.
            - Only considers major.minor format version, ignores patch version.
            - Conversion to lower format versions is not supported.
        context: Validation context, see `bioimagieo.spec.ValidationContext`

    Note:
        Use `bioimagieo.spec.load_description_and_validate_format_only` to validate a
        file source instead of loading the YAML content and creating the appropriate
        `ValidationContext`.

        Alternatively you can use `bioimagieo.spec.load_description` and access the
        `validation_summary` attribute of the returned object.
    """
    with context or get_validation_context():
        rd = build_description(data, format_version=format_version)

    assert rd.validation_summary is not None
    return rd.validation_summary


def ensure_description_is_model(
    rd: Union[InvalidDescr, ResourceDescr],
) -> AnyModelDescr:
    """
    Raises:
        ValueError: for invalid or non-model resources
    """
    if isinstance(rd, InvalidDescr):
        rd.validation_summary.display()
        raise ValueError(f"Invalid {rd.type} description")

    if rd.type != "model":
        rd.validation_summary.display()
        raise ValueError(
            f"Expected a model resource, but got resource type '{rd.type}'"
        )

    assert not isinstance(
        rd,
        (
            GenericDescr_v0_2,
            GenericDescr_v0_3,
        ),
    )

    return rd


def ensure_description_is_dataset(
    rd: Union[InvalidDescr, ResourceDescr],
) -> AnyDatasetDescr:
    if isinstance(rd, InvalidDescr):
        rd.validation_summary.display()
        raise ValueError(f"Invalid {rd.type} description.")

    if rd.type != "dataset":
        rd.validation_summary.display()
        raise ValueError(
            f"Expected a dataset resource, but got resource type '{rd.type}'"
        )

    assert not isinstance(
        rd,
        (
            GenericDescr_v0_2,
            GenericDescr_v0_3,
        ),
    )

    return rd



from __future__ import annotations

from datetime import date, datetime
from typing import (
    Any,
    Hashable,
    Mapping,
    Sequence,
    Union,
)

import httpx

from ._settings import settings
from .constants import KNOWN_GITHUB_USERS, KNOWN_INVALID_GITHUB_USERS
from .field_warning import issue_warning
from .type_guards import is_mapping, is_sequence, is_tuple
from .validation_context import get_validation_context


def is_valid_yaml_leaf_value(value: Any) -> bool:
    return value is None or isinstance(value, (bool, date, datetime, int, float, str))


def is_valid_yaml_key(value: Union[Any, Sequence[Any]]) -> bool:
    return (
        is_valid_yaml_leaf_value(value)
        or is_tuple(value)
        and all(is_valid_yaml_leaf_value(v) for v in value)
    )


def is_valid_yaml_mapping(value: Union[Any, Mapping[Any, Any]]) -> bool:
    return is_mapping(value) and all(
        is_valid_yaml_key(k) and is_valid_yaml_value(v) for k, v in value.items()
    )


def is_valid_yaml_sequence(value: Union[Any, Sequence[Any]]) -> bool:
    return is_sequence(value) and all(is_valid_yaml_value(v) for v in value)


def is_valid_yaml_value(value: Any) -> bool:
    return any(
        is_valid(value)
        for is_valid in (
            is_valid_yaml_key,
            is_valid_yaml_mapping,
            is_valid_yaml_sequence,
        )
    )


def validate_unique_entries(seq: Sequence[Hashable]):
    if len(seq) != len(set(seq)):
        raise ValueError("Entries are not unique.")
    return seq


def validate_github_user(
    username: str, hotfix_known_errorenous_names: bool = True
) -> str:
    if hotfix_known_errorenous_names:
        if username == "Constantin Pape":
            return "constantinpape"

    if (
        username.lower() in KNOWN_GITHUB_USERS
        or not get_validation_context().perform_io_checks
    ):
        return username

    if username.lower() in KNOWN_INVALID_GITHUB_USERS:
        raise ValueError(f"Known invalid GitHub user '{username}'")

    try:
        r = httpx.get(
            f"https://api.github.com/users/{username}",
            auth=settings.github_auth,
            timeout=settings.http_timeout,
        )
    except httpx.TimeoutException:
        issue_warning(
            "Could not verify GitHub user '{value}' due to connection timeout",
            value=username,
        )
    else:
        if r.status_code == 403 and r.reason_phrase == "rate limit exceeded":
            issue_warning(
                "Could not verify GitHub user '{value}' due to GitHub API rate limit",
                value=username,
            )
        elif r.status_code != 200:
            KNOWN_INVALID_GITHUB_USERS.add(username.lower())
            raise ValueError(f"Could not find GitHub user '{username}'")

        KNOWN_GITHUB_USERS.add(username.lower())

    return username



import difflib
import re
from typing import Any, Dict, Optional, Tuple

from loguru import logger
from pydantic import ConfigDict

from .._description import DESCRIPTIONS_MAP, SpecificResourceDescr
from .._version import VERSION
from .type_guards import is_dict, is_list


def _patch_desccription(
    json_schema: Dict[str, Any], path: Tuple[str, ...] = ()
) -> None:
    """Patch descriptions:

    - replace mkdocstrings cross-reference syntax [object][] by `object`
    """
    if isinstance(descr := json_schema.get("description"), str) and descr:
        # Replace markdown link references like [object][] with `object`
        new_descr = re.sub(r"\[([^\]]+)\]\[\]", r"`\1`", descr)
        if descr != new_descr:
            diff = difflib.Differ().compare(descr.splitlines(), new_descr.splitlines())
            logger.debug(f"updated {'.'.join(path)}:\n" + "\n".join(diff))
            json_schema["description"] = new_descr

    for k, v in json_schema.items():
        if is_dict(v):
            _patch_desccription(v, path + (str(k),))
        elif is_list(v):
            for index, item in enumerate(v):
                if is_dict(item):
                    _patch_desccription(item, path + (str(k), str(index)))


def generate_json_schema(
    type_format: Optional[Tuple[str, str]] = None,
) -> Dict[str, Any]:
    """generate the bioimageio specification as a JSON schema"""
    from pydantic import TypeAdapter

    spec_format_version = ".".join(VERSION.split(".")[:3])  # strip library version
    if type_format is None:
        adapter: TypeAdapter[Any] = TypeAdapter(
            SpecificResourceDescr,
            config=ConfigDict(
                title=f"bioimage.io resource description {spec_format_version}"
            ),
        )
    else:
        typ, format_version = type_format
        rd_class = DESCRIPTIONS_MAP[typ][format_version]
        adapter = TypeAdapter(rd_class)

    json_schema = adapter.json_schema()
    _patch_desccription(json_schema)
    return json_schema



from __future__ import annotations

import json
from types import MappingProxyType
from typing import Mapping, NamedTuple, Sequence, Set, Union

from .utils import files

DOI_REGEX = (  # lax DOI regex validating the first 7 DOI characters only
    r"^10\.[0-9]{4}.+$"
)

SHA256_HINT = """You can drag and drop your file to this
[online tool](http://emn178.github.io/online-tools/sha256_checksum.html) to generate a SHA256 in your browser.
Or you can generate a SHA256 checksum with Python's `hashlib`,
[here is a codesnippet](https://gist.github.com/FynnBe/e64460463df89439cff218bbf59c1100)."""

with (
    files("bioimageio.spec")
    .joinpath("static/tag_categories.json")
    .open("r", encoding="utf-8")
) as f:
    TAG_CATEGORIES: Mapping[str, Mapping[str, Sequence[str]]] = json.load(f)

# SI unit regex adapted from https://stackoverflow.com/a/3573731
_prefix = "(Q|R|Y|Z|E|P|T|G|M|k|h|da|d|c|m|µ|n|p|f|a|z|y|r|q)"
_unit = "(m|g|s|A|K|mol|cd|Hz|N|Pa|J|W|C|V|F|Ω|S|Wb|T|H|lm|lx|Bq|Gy|Sv|kat|l|L)"
_any_power = r"(\^[+-]?[1-9]\d*)"
_pos_power = r"(\^+?[1-9]\d*)"
_unit_ap = f"{_prefix}?{_unit}{_any_power}?"
_unit_pp = f"{_prefix}?{_unit}{_pos_power}?"
SI_UNIT_REGEX = f"^{_unit_ap}((·{_unit_ap})|(/{_unit_pp}))*$"


class _DtypeLimit(NamedTuple):
    min: Union[int, float]
    max: Union[int, float]


# numpy.dtype limits; see scripts/generate_dtype_limits.py
DTYPE_LIMITS = MappingProxyType(
    {
        "float32": _DtypeLimit(-3.4028235e38, 3.4028235e38),
        "float64": _DtypeLimit(-1.7976931348623157e308, 1.7976931348623157e308),
        "uint8": _DtypeLimit(0, 255),
        "int8": _DtypeLimit(-128, 127),
        "uint16": _DtypeLimit(0, 65535),
        "int16": _DtypeLimit(-32768, 32767),
        "uint32": _DtypeLimit(0, 4294967295),
        "int32": _DtypeLimit(-2147483648, 2147483647),
        "uint64": _DtypeLimit(0, 18446744073709551615),
        "int64": _DtypeLimit(-9223372036854775808, 9223372036854775807),
    }
)

# TODO: cache/store known GitHub users in file
KNOWN_GITHUB_USERS: Set[str] = {
    "aaitorg",
    "anwai98",
    "bioimageiobot",
    "carlosuc3m",
    "cfusterbarcelo",
    "clementcaporal",
    "constantinpape",
    "ctr26",
    "danifranco",
    "donglaiw",
    "esgomezm",
    "fynnbe",
    "githubuser2",
    "iarganda",
    "ilastik",
    "ivanhcenalmor",
    "jansanrom",
    "k-dominik",
    "lenkaback",
    "oeway",
    "pedgomgal1",
}

N_KNOWN_GITHUB_USERS = len(KNOWN_GITHUB_USERS)
KNOWN_INVALID_GITHUB_USERS: Set[str] = {"arratemunoz", "lmescu"}
N_KNOWN_INVALID_GITHUB_USERS = len(KNOWN_INVALID_GITHUB_USERS)



from typing import Literal

WarningSeverity = Literal[20, 30, 35]
WarningLevel = Literal[WarningSeverity, 50]
"""With warning level x validation warnings of severity >=x are raised.
Highest warning level 50/error does not raise any validaiton warnings (only validation errors)."""

ERROR, ERROR_NAME = 50, "error"
"""A warning of the error level is always raised (equivalent to a validation error)"""

ALERT, ALERT_NAME = 35, "alert"
"""no ALERT nor ERROR -> RDF is worriless"""

WARNING, WARNING_NAME = 30, "warning"
"""no WARNING nor ALERT nor ERROR -> RDF is watertight"""

INFO, INFO_NAME = 20, "info"
"""info warnings are about purely cosmetic issues, etc."""



from typing import TYPE_CHECKING, Any, ClassVar, Dict, Literal, Optional, cast

from pydantic import model_validator

from .._internal.common_nodes import InvalidDescr
from .._internal.io import FileDescr as FileDescr
from .._internal.io_basics import Sha256 as Sha256
from .._internal.types import FAIR
from .._internal.url import HttpUrl as HttpUrl
from ..generic.v0_3 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_3 import Author as Author
from ..generic.v0_3 import BadgeDescr as BadgeDescr
from ..generic.v0_3 import BioimageioConfig as BioimageioConfig
from ..generic.v0_3 import CiteEntry as CiteEntry
from ..generic.v0_3 import Config as Config
from ..generic.v0_3 import DeprecatedLicenseId as DeprecatedLicenseId
from ..generic.v0_3 import Doi as Doi
from ..generic.v0_3 import (
    GenericDescrBase,
    LinkedResourceBase,
    _author_conv,  # pyright: ignore[reportPrivateUsage]
    _maintainer_conv,  # pyright: ignore[reportPrivateUsage]
)
from ..generic.v0_3 import LicenseId as LicenseId
from ..generic.v0_3 import LinkedResource as LinkedResource
from ..generic.v0_3 import Maintainer as Maintainer
from ..generic.v0_3 import OrcidId as OrcidId
from ..generic.v0_3 import RelativeFilePath as RelativeFilePath
from ..generic.v0_3 import ResourceId as ResourceId
from ..generic.v0_3 import Uploader as Uploader
from ..generic.v0_3 import Version as Version
from .v0_2 import DatasetDescr as DatasetDescr02


class DatasetId(ResourceId):
    pass


class DatasetDescr(GenericDescrBase):
    """A bioimage.io dataset resource description file (dataset RDF) describes a dataset relevant to bioimage
    processing.
    """

    implemented_type: ClassVar[Literal["dataset"]] = "dataset"
    if TYPE_CHECKING:
        type: Literal["dataset"] = "dataset"
    else:
        type: Literal["dataset"]

    id: Optional[DatasetId] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    parent: Optional[DatasetId] = None
    """The description from which this one is derived"""

    source: FAIR[Optional[HttpUrl]] = None
    """"URL to the source of the dataset."""

    @model_validator(mode="before")
    @classmethod
    def _convert(cls, data: Dict[str, Any], /) -> Dict[str, Any]:
        if (
            data.get("type") == "dataset"
            and isinstance(fv := data.get("format_version"), str)
            and fv.startswith("0.2.")
        ):
            old = DatasetDescr02.load(data)
            if isinstance(old, InvalidDescr):
                return data

            return cast(
                Dict[str, Any],
                (cls if TYPE_CHECKING else dict)(
                    attachments=(
                        []
                        if old.attachments is None
                        else [FileDescr(source=f) for f in old.attachments.files]
                    ),
                    authors=[_author_conv.convert_as_dict(a) for a in old.authors],  # pyright: ignore[reportArgumentType]
                    badges=old.badges,
                    cite=[
                        {"text": c.text, "doi": c.doi, "url": c.url} for c in old.cite
                    ],  # pyright: ignore[reportArgumentType]
                    config=old.config,  # pyright: ignore[reportArgumentType]
                    covers=old.covers,
                    description=old.description,
                    documentation=old.documentation,
                    format_version="0.3.0",
                    git_repo=old.git_repo,  # pyright: ignore[reportArgumentType]
                    icon=old.icon,
                    id=None if old.id is None else DatasetId(old.id),
                    license=old.license,  # type: ignore
                    links=old.links,
                    maintainers=[
                        _maintainer_conv.convert_as_dict(m) for m in old.maintainers
                    ],  # pyright: ignore[reportArgumentType]
                    name=old.name,
                    source=old.source,
                    tags=old.tags,
                    type=old.type,
                    uploader=old.uploader,
                    version=old.version,
                    **(old.model_extra or {}),
                ),
            )

        return data


class LinkedDataset(LinkedResourceBase):
    """Reference to a bioimage.io dataset."""

    id: DatasetId
    """A valid dataset `id` from the bioimage.io collection."""



from __future__ import annotations

import string
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import annotated_types
from annotated_types import Len, LowerCase, MaxLen, MinLen
from pydantic import Field, RootModel, ValidationInfo, field_validator, model_validator
from typing_extensions import Annotated

from .._internal.common_nodes import Node, ResourceDescrBase
from .._internal.constants import TAG_CATEGORIES
from .._internal.field_validation import validate_github_user
from .._internal.field_warning import as_warning, issue_warning, warn
from .._internal.io import (
    BioimageioYamlContent,
    FileDescr,
    WithSuffix,
    is_yaml_value,
)
from .._internal.io_basics import Sha256
from .._internal.io_packaging import FileDescr_
from .._internal.license_id import DeprecatedLicenseId, LicenseId
from .._internal.node_converter import Converter
from .._internal.type_guards import is_dict
from .._internal.types import FAIR, FileSource_, NotEmpty, RelativeFilePath
from .._internal.url import HttpUrl
from .._internal.validated_string import ValidatedString
from .._internal.validator_annotations import (
    Predicate,
    RestrictCharacters,
)
from .._internal.version_type import Version
from .._internal.warning_levels import ALERT, INFO
from ._v0_3_converter import convert_from_older_format
from .v0_2 import Author as _Author_v0_2
from .v0_2 import BadgeDescr, Doi, FileSource_cover, OrcidId, Uploader
from .v0_2 import Maintainer as _Maintainer_v0_2

__all__ = [
    "Author",
    "BadgeDescr",
    "CiteEntry",
    "DeprecatedLicenseId",
    "Doi",
    "FileDescr",
    "GenericDescr",
    "HttpUrl",
    "KNOWN_SPECIFIC_RESOURCE_TYPES",
    "LicenseId",
    "LinkedResource",
    "Maintainer",
    "OrcidId",
    "RelativeFilePath",
    "ResourceId",
    "Sha256",
    "Uploader",
    "VALID_COVER_IMAGE_EXTENSIONS",
    "Version",
]

KNOWN_SPECIFIC_RESOURCE_TYPES = (
    "application",
    "collection",
    "dataset",
    "model",
    "notebook",
)
VALID_COVER_IMAGE_EXTENSIONS = (
    ".gif",
    ".jpeg",
    ".jpg",
    ".png",
    ".svg",
)


class ResourceId(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            NotEmpty[str],
            RestrictCharacters(string.ascii_lowercase + string.digits + "_-/."),
            annotated_types.Predicate(
                lambda s: not (s.startswith("/") or s.endswith("/"))
            ),
        ]
    ]


def _has_no_slash(s: str) -> bool:
    return "/" not in s and "\\" not in s


class Author(_Author_v0_2):
    name: Annotated[str, Predicate(_has_no_slash)]
    github_user: Optional[str] = None

    @field_validator("github_user", mode="after")
    def _validate_github_user(cls, value: Optional[str]):
        if value is None:
            return None
        else:
            return validate_github_user(value)


class _AuthorConv(Converter[_Author_v0_2, Author]):
    def _convert(
        self, src: _Author_v0_2, tgt: "type[Author] | type[dict[str, Any]]"
    ) -> "Author | dict[str, Any]":
        return tgt(
            name=src.name,
            github_user=src.github_user,
            affiliation=src.affiliation,
            email=src.email,
            orcid=src.orcid,
        )


_author_conv = _AuthorConv(_Author_v0_2, Author)


class Maintainer(_Maintainer_v0_2):
    name: Optional[Annotated[str, Predicate(_has_no_slash)]] = None
    github_user: str

    @field_validator("github_user", mode="after")
    def validate_github_user(cls, value: str):
        return validate_github_user(value)


class _MaintainerConv(Converter[_Maintainer_v0_2, Maintainer]):
    def _convert(
        self, src: _Maintainer_v0_2, tgt: "type[Maintainer | dict[str, Any]]"
    ) -> "Maintainer | dict[str, Any]":
        return tgt(
            name=src.name,
            github_user=src.github_user,
            affiliation=src.affiliation,
            email=src.email,
            orcid=src.orcid,
        )


_maintainer_conv = _MaintainerConv(_Maintainer_v0_2, Maintainer)


class CiteEntry(Node):
    """A citation that should be referenced in work using this resource."""

    text: str
    """free text description"""

    doi: Optional[Doi] = None
    """A digital object identifier (DOI) is the prefered citation reference.
    See https://www.doi.org/ for details.
    Note:
        Either **doi** or **url** have to be specified.
    """

    url: Optional[HttpUrl] = None
    """URL to cite (preferably specify a **doi** instead/also).
    Note:
        Either **doi** or **url** have to be specified.
    """

    @model_validator(mode="after")
    def _check_doi_or_url(self):
        if not self.doi and not self.url:
            raise ValueError("Either 'doi' or 'url' is required")

        return self


class LinkedResourceBase(Node):
    @model_validator(mode="before")
    def _remove_version_number(cls, value: Any):
        if is_dict(value):
            vn = value.pop("version_number", None)
            if vn is not None and value.get("version") is None:
                value["version"] = vn

        return value

    version: Optional[Version] = None
    """The version of the linked resource following SemVer 2.0."""


class LinkedResource(LinkedResourceBase):
    """Reference to a bioimage.io resource"""

    id: ResourceId
    """A valid resource `id` from the official bioimage.io collection."""


class BioimageioConfig(Node, extra="allow"):
    """bioimage.io internal metadata."""


class Config(Node, extra="allow"):
    """A place to store additional metadata (often tool specific).

    Such additional metadata is typically set programmatically by the respective tool
    or by people with specific insights into the tool.
    If you want to store additional metadata that does not match any of the other
    fields, think of a key unlikely to collide with anyone elses use-case/tool and save
    it here.

    Please consider creating [an issue in the bioimageio.spec repository](https://github.com/bioimage-io/spec-bioimage-io/issues/new?template=Blank+issue)
    if you are not sure if an existing field could cover your use case
    or if you think such a field should exist.
    """

    bioimageio: BioimageioConfig = Field(default_factory=BioimageioConfig)
    """bioimage.io internal metadata."""

    @model_validator(mode="after")
    def _validate_extra_fields(self):
        if self.model_extra:
            for k, v in self.model_extra.items():
                if not isinstance(v, Node) and not is_yaml_value(v):
                    raise ValueError(
                        f"config.{k} is not a valid YAML value or `Node` instance"
                    )

        return self

    def __getitem__(self, key: str) -> Any:
        """Allows to access the config as a dictionary."""
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        """Allows to set the config as a dictionary."""
        setattr(self, key, value)


class GenericModelDescrBase(ResourceDescrBase):
    """Base for all resource descriptions including of model descriptions"""

    name: Annotated[
        Annotated[
            str, RestrictCharacters(string.ascii_letters + string.digits + "_+- ()")
        ],
        MinLen(5),
        MaxLen(128),
        warn(MaxLen(64), "Name longer than 64 characters.", INFO),
    ]
    """A human-friendly name of the resource description.
    May only contains letters, digits, underscore, minus, parentheses and spaces."""

    description: FAIR[
        Annotated[
            str,
            MaxLen(1024),
            warn(MaxLen(512), "Description longer than 512 characters."),
        ]
    ] = ""
    """A string containing a brief description."""

    covers: List[FileSource_cover] = Field(
        default_factory=cast(Callable[[], List[FileSource_cover]], list),
        description=(
            "Cover images. Please use an image smaller than 500KB and an aspect"
            " ratio width to height of 2:1 or 1:1.\nThe supported image formats"
            f" are: {VALID_COVER_IMAGE_EXTENSIONS}"
        ),
        examples=[["cover.png"]],
    )
    """Cover images."""

    id_emoji: Optional[
        Annotated[str, Len(min_length=1, max_length=2), Field(examples=["🦈", "🦥"])]
    ] = None
    """UTF-8 emoji for display alongside the `id`."""

    authors: FAIR[List[Author]] = Field(
        default_factory=cast(Callable[[], List[Author]], list)
    )
    """The authors are the creators of this resource description and the primary points of contact."""

    attachments: List[FileDescr_] = Field(
        default_factory=cast(Callable[[], List[FileDescr_]], list)
    )
    """file attachments"""

    cite: FAIR[List[CiteEntry]] = Field(
        default_factory=cast(Callable[[], List[CiteEntry]], list)
    )
    """citations"""

    license: FAIR[
        Annotated[
            Annotated[
                Union[LicenseId, DeprecatedLicenseId, None],
                Field(union_mode="left_to_right"),
            ],
            warn(
                Optional[LicenseId],
                "{value} is deprecated, see https://spdx.org/licenses/{value}.html",
            ),
            Field(examples=["CC0-1.0", "MIT", "BSD-2-Clause"]),
        ]
    ] = None
    """A [SPDX license identifier](https://spdx.org/licenses/).
    We do not support custom license beyond the SPDX license list, if you need that please
    [open a GitHub issue](https://github.com/bioimage-io/spec-bioimage-io/issues/new/choose)
    to discuss your intentions with the community."""

    git_repo: Annotated[
        Optional[HttpUrl],
        Field(
            examples=[
                "https://github.com/bioimage-io/spec-bioimage-io/tree/main/example_descriptions/models/unet2d_nuclei_broad"
            ],
        ),
    ] = None
    """A URL to the Git repository where the resource is being developed."""

    icon: Union[Annotated[str, Len(min_length=1, max_length=2)], FileSource_, None] = (
        None
    )
    """An icon for illustration, e.g. on bioimage.io"""

    links: Annotated[
        List[str],
        Field(
            examples=[
                (
                    "ilastik/ilastik",
                    "deepimagej/deepimagej",
                    "zero/notebook_u-net_3d_zerocostdl4mic",
                )
            ],
        ),
    ] = Field(default_factory=list)
    """IDs of other bioimage.io resources"""

    uploader: Optional[Uploader] = None
    """The person who uploaded the model (e.g. to bioimage.io)"""

    maintainers: List[Maintainer] = Field(
        default_factory=cast(Callable[[], List[Maintainer]], list)
    )
    """Maintainers of this resource.
    If not specified, `authors` are maintainers and at least some of them has to specify their `github_user` name"""

    @model_validator(mode="after")
    def _check_maintainers_exist(self):
        if not self.maintainers and self.authors:
            if all(a.github_user is None for a in self.authors):
                issue_warning(
                    "Missing `maintainers` or any author in `authors` with a specified"
                    + " `github_user` name.",
                    value=self.authors,
                    field="authors",
                    severity=ALERT,
                )

        return self

    tags: FAIR[
        Annotated[
            List[str],
            Field(
                examples=[("unet2d", "pytorch", "nucleus", "segmentation", "dsb2018")]
            ),
        ]
    ] = Field(default_factory=list)
    """Associated tags"""

    @as_warning
    @field_validator("tags")
    @classmethod
    def warn_about_tag_categories(
        cls, value: List[str], info: ValidationInfo
    ) -> List[str]:
        categories = TAG_CATEGORIES.get(info.data["type"], {})
        missing_categories: List[Dict[str, Sequence[str]]] = []
        for cat, entries in categories.items():
            if not any(e in value for e in entries):
                missing_categories.append({cat: entries})

        if missing_categories:
            raise ValueError(
                f"Missing tags from bioimage.io categories: {missing_categories}"
            )

        return value

    version: Optional[Version] = None
    """The version of the resource following SemVer 2.0."""

    @model_validator(mode="before")
    def _remove_version_number(cls, value: Any):
        if is_dict(value):
            vn = value.pop("version_number", None)
            if vn is not None and value.get("version") is None:
                value["version"] = vn

        return value

    version_comment: Optional[Annotated[str, MaxLen(512)]] = None
    """A comment on the version of the resource."""


FileSource_documentation = Annotated[
    FileSource_,
    WithSuffix(".md", case_sensitive=True),
    Field(
        examples=[
            "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_descriptions/models/unet2d_nuclei_broad/README.md",
            "README.md",
        ],
    ),
]


class GenericDescrBase(GenericModelDescrBase):
    """Base for all resource descriptions except for the model descriptions"""

    implemented_format_version: ClassVar[Literal["0.3.0"]] = "0.3.0"
    if TYPE_CHECKING:
        format_version: Literal["0.3.0"] = "0.3.0"
    else:
        format_version: Literal["0.3.0"]
        """The **format** version of this resource specification"""

    @model_validator(mode="before")
    @classmethod
    def _convert_from_older_format(
        cls, data: BioimageioYamlContent, /
    ) -> BioimageioYamlContent:
        cls.convert_from_old_format_wo_validation(data)
        return data

    @classmethod
    def convert_from_old_format_wo_validation(cls, data: BioimageioYamlContent) -> None:
        """Convert metadata following an older format version to this classes' format
        without validating the result.
        """
        convert_from_older_format(data)

    documentation: FAIR[Optional[FileSource_documentation]] = None
    """URL or relative path to a markdown file encoded in UTF-8 with additional documentation.
    The recommended documentation file name is `README.md`. An `.md` suffix is mandatory."""

    badges: List[BadgeDescr] = Field(  # pyright: ignore[reportUnknownVariableType]
        default_factory=list
    )
    """badges associated with this resource"""

    config: Config = Field(default_factory=Config.model_construct)
    """A field for custom configuration that can contain any keys not present in the RDF spec.
    This means you should not store, for example, a GitHub repo URL in `config` since there is a `git_repo` field.
    Keys in `config` may be very specific to a tool or consumer software. To avoid conflicting definitions,
    it is recommended to wrap added configuration into a sub-field named with the specific domain or tool name,
    for example:
    ```yaml
    config:
        giraffe_neckometer:  # here is the domain name
            length: 3837283
            address:
                home: zoo
        imagej:              # config specific to ImageJ
            macro_dir: path/to/macro/file
    ```
    If possible, please use [`snake_case`](https://en.wikipedia.org/wiki/Snake_case) for keys in `config`.
    You may want to list linked files additionally under `attachments` to include them when packaging a resource.
    (Packaging a resource means downloading/copying important linked files and creating a ZIP archive that contains
    an altered rdf.yaml file with local references to the downloaded files.)"""


ResourceDescrType = TypeVar("ResourceDescrType", bound=GenericDescrBase)


class GenericDescr(GenericDescrBase, extra="ignore"):
    """Specification of the fields used in a generic bioimage.io-compliant resource description file (RDF).

    An RDF is a YAML file that describes a resource such as a model, a dataset, or a notebook.
    Note that those resources are described with a type-specific RDF.
    Use this generic resource description, if none of the known specific types matches your resource.
    """

    implemented_type: ClassVar[Literal["generic"]] = "generic"
    if TYPE_CHECKING:
        type: Annotated[str, LowerCase] = "generic"
        """The resource type assigns a broad category to the resource."""
    else:
        type: Annotated[str, LowerCase]
        """The resource type assigns a broad category to the resource."""

    id: Optional[
        Annotated[ResourceId, Field(examples=["affable-shark", "ambitious-sloth"])]
    ] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    parent: Optional[ResourceId] = None
    """The description from which this one is derived"""

    source: Optional[HttpUrl] = None
    """The primary source of the resource"""

    @field_validator("type", mode="after")
    @classmethod
    def check_specific_types(cls, value: str) -> str:
        if value in KNOWN_SPECIFIC_RESOURCE_TYPES:
            raise ValueError(
                f"Use the {value} description instead of this generic description for"
                + f" your '{value}' resource."
            )

        return value



from typing import TYPE_CHECKING, ClassVar, Literal, Optional

from pydantic import Field
from typing_extensions import Annotated

from .._internal.common_nodes import Node
from .._internal.types import FileSource_ as FileSource_
from .._internal.url import HttpUrl as HttpUrl
from ..generic.v0_2 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_2 import AttachmentsDescr as AttachmentsDescr
from ..generic.v0_2 import Author as Author
from ..generic.v0_2 import BadgeDescr as BadgeDescr
from ..generic.v0_2 import CiteEntry as CiteEntry
from ..generic.v0_2 import Doi as Doi
from ..generic.v0_2 import GenericDescrBase
from ..generic.v0_2 import LinkedResource as LinkedResource
from ..generic.v0_2 import Maintainer as Maintainer
from ..generic.v0_2 import OrcidId as OrcidId
from ..generic.v0_2 import RelativeFilePath as RelativeFilePath
from ..generic.v0_2 import ResourceId as ResourceId
from ..generic.v0_2 import Uploader as Uploader
from ..generic.v0_2 import Version as Version


class ApplicationId(ResourceId):
    pass


class ApplicationDescr(GenericDescrBase):
    """Bioimage.io description of an application."""

    implemented_type: ClassVar[Literal["application"]] = "application"
    if TYPE_CHECKING:
        type: Literal["application"] = "application"
    else:
        type: Literal["application"]

    id: Optional[ApplicationId] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    source: Annotated[
        Optional[FileSource_],
        Field(description="URL or path to the source of the application"),
    ] = None
    """The primary source of the application"""


class LinkedApplication(Node):
    """Reference to a bioimage.io application."""

    id: ApplicationId
    """A valid application `id` from the bioimage.io collection."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version) of linked application"""



"""Common nodes and field types of bioimage.io specifications."""

from pydantic import ValidationError

from ._internal.common_nodes import InvalidDescr
from ._internal.io import (
    BioimageioYamlContent,
    BioimageioYamlSource,
    FileDescr,
    YamlValue,
)
from ._internal.io_basics import (
    AbsoluteDirectory,
    AbsoluteFilePath,
    BytesReader,
    FileName,
    Sha256,
    ZipPath,
)
from ._internal.root_url import RootHttpUrl
from ._internal.types import (
    FilePath,
    FileSource,
    PermissiveFileSource,
    RelativeFilePath,
)
from ._internal.url import HttpUrl

__all__ = [
    "AbsoluteDirectory",
    "AbsoluteFilePath",
    "BioimageioYamlContent",
    "BioimageioYamlSource",
    "BytesReader",
    "FileDescr",
    "FileName",
    "FilePath",
    "FileSource",
    "HttpUrl",
    "InvalidDescr",
    "PermissiveFileSource",
    "RelativeFilePath",
    "RootHttpUrl",
    "Sha256",
    "ValidationError",
    "YamlValue",
    "ZipPath",
]



from typing import Any, Optional, Protocol

from rich.progress import Progress


class Progressbar(Protocol):
    """Progressbar protocol modeled after tqdm"""

    total: Optional[int]

    def update(self, increment: int, /) -> Any: ...

    def reset(self): ...

    def close(self): ...

    def set_description(self, description: str, /, refresh: bool = True): ...


class RichTaskBar:
    def __init__(self, description: str, *, parent: Progress, total: Optional[int]):
        super().__init__()
        self.task_id = parent.add_task(description, total=total)
        self.parent = parent
        self.total = total

    def update(self, increment: int, /):
        self.parent.advance(self.task_id, increment)

    def reset(self):
        self.parent.reset(self.task_id)

    def close(self):
        self.parent.remove_task(self.task_id)

    def set_description(self, description: str, /, refresh: bool = True):
        self.parent.update(self.task_id, description=description, refresh=refresh)


class RichOverallProgress:
    def __init__(self):
        super().__init__()
        self.progress = Progress()

    def __call__(self, description: str = "", *, total: Optional[int] = None):
        return RichTaskBar(description, parent=self.progress, total=total)



from typing import List, Literal, Optional, Union

from typing_extensions import assert_never

from ._internal.gh_utils import set_github_warning
from ._internal.io import FileDescr, get_reader
from ._internal.io_utils import read_yaml
from .conda_env import BioimageioCondaEnv, PipDeps
from .model import v0_4, v0_5
from .model.v0_5 import Version

SupportedWeightsEntry = Union[
    v0_4.KerasHdf5WeightsDescr,
    v0_4.OnnxWeightsDescr,
    v0_4.PytorchStateDictWeightsDescr,
    v0_4.TensorflowSavedModelBundleWeightsDescr,
    v0_4.TorchscriptWeightsDescr,
    v0_5.KerasHdf5WeightsDescr,
    v0_5.OnnxWeightsDescr,
    v0_5.PytorchStateDictWeightsDescr,
    v0_5.TensorflowSavedModelBundleWeightsDescr,
    v0_5.TorchscriptWeightsDescr,
]


def get_conda_env(
    *,
    entry: SupportedWeightsEntry,
    env_name: Optional[Union[Literal["DROP"], str]] = None,
) -> BioimageioCondaEnv:
    """get the recommended Conda environment for a given weights entry description"""
    if isinstance(entry, (v0_4.OnnxWeightsDescr, v0_5.OnnxWeightsDescr)):
        conda_env = _get_default_onnx_env(opset_version=entry.opset_version)
    elif isinstance(
        entry,
        (
            v0_4.PytorchStateDictWeightsDescr,
            v0_5.PytorchStateDictWeightsDescr,
            v0_4.TorchscriptWeightsDescr,
            v0_5.TorchscriptWeightsDescr,
        ),
    ):
        if (
            isinstance(entry, v0_5.TorchscriptWeightsDescr)
            or entry.dependencies is None
        ):
            conda_env = _get_default_pytorch_env(pytorch_version=entry.pytorch_version)
        else:
            conda_env = _get_env_from_deps(entry.dependencies)

    elif isinstance(
        entry,
        (
            v0_4.TensorflowSavedModelBundleWeightsDescr,
            v0_5.TensorflowSavedModelBundleWeightsDescr,
        ),
    ):
        if entry.dependencies is None:
            conda_env = _get_default_tf_env(tensorflow_version=entry.tensorflow_version)
        else:
            conda_env = _get_env_from_deps(entry.dependencies)
    elif isinstance(
        entry,
        (v0_4.KerasHdf5WeightsDescr, v0_5.KerasHdf5WeightsDescr),
    ):
        conda_env = _get_default_tf_env(tensorflow_version=entry.tensorflow_version)
    else:
        assert_never(entry)

    if env_name == "DROP":
        conda_env.name = None
    elif env_name is not None:
        conda_env.name = env_name

    return conda_env


def _get_default_pytorch_env(
    *,
    pytorch_version: Optional[Version] = None,
) -> BioimageioCondaEnv:
    if pytorch_version is None:
        pytorch_version = Version("1.10.1")

    channels = ["conda-forge", "nodefaults"]
    if pytorch_version < Version("2.5.2"):
        channels.insert(0, "pytorch")

    # dependencies to install pytorch according to
    # https://pytorch.org/get-started/previous-versions/
    v = pytorch_version.base_version
    if v.count(".") == 0:
        v += ".0.0"
    elif v.count(".") == 1:
        v += ".0"

    deps: List[Union[str, PipDeps]] = [f"pytorch=={v}"]
    if v == "1.5.1":
        deps += ["torchvision==0.6.1"]
    elif v == "1.6.0":
        deps += ["torchvision==0.7.0"]
    elif v == "1.7.0":
        deps += ["torchvision==0.8.0", "torchaudio==0.7.0"]
    elif v == "1.7.1":
        deps += ["torchvision==0.8.2", "torchaudio==0.7.1"]
    elif v == "1.8.0":
        deps += ["torchvision==0.9.0", "torchaudio==0.8.0"]
    elif v == "1.8.1":
        deps += ["torchvision==0.9.1", "torchaudio==0.8.1"]
    elif v == "1.9.0":
        deps += ["torchvision==0.10.0", "torchaudio==0.9.0"]
    elif v == "1.9.1":
        deps += ["torchvision==0.10.1", "torchaudio==0.9.1"]
    elif v == "1.10.0":
        deps += ["torchvision==0.11.0", "torchaudio==0.10.0"]
    elif v == "1.10.1":
        deps += ["torchvision==0.11.2", "torchaudio==0.10.1"]
    elif v == "1.11.0":
        deps += ["torchvision==0.12.0", "torchaudio==0.11.0"]
    elif v == "1.12.0":
        deps += ["torchvision==0.13.0", "torchaudio==0.12.0"]
    elif v == "1.12.1":
        deps += ["torchvision==0.13.1", "torchaudio==0.12.1"]
    elif v == "1.13.0":
        deps += ["torchvision==0.14.0", "torchaudio==0.13.0"]
    elif v == "1.13.1":
        deps += ["torchvision==0.14.1", "torchaudio==0.13.1"]
    elif v == "2.0.0":
        deps += ["torchvision==0.15.0", "torchaudio==2.0.0"]
    elif v == "2.0.1":
        deps += ["torchvision==0.15.2", "torchaudio==2.0.2"]
    elif v == "2.1.0":
        deps += ["torchvision==0.16.0", "torchaudio==2.1.0"]
    elif v == "2.1.1":
        deps += ["torchvision==0.16.1", "torchaudio==2.1.1"]
    elif v == "2.1.2":
        deps += ["torchvision==0.16.2", "torchaudio==2.1.2"]
    elif v == "2.2.0":
        deps += ["torchvision==0.17.0", "torchaudio==2.2.0"]
    elif v == "2.2.1":
        deps += ["torchvision==0.17.1", "torchaudio==2.2.1"]
    elif v == "2.2.2":
        deps += ["torchvision==0.17.2", "torchaudio==2.2.2"]
    elif v == "2.3.0":
        deps += ["torchvision==0.18.0", "torchaudio==2.3.0"]
    elif v == "2.3.1":
        deps += ["torchvision==0.18.1", "torchaudio==2.3.1"]
    elif v == "2.4.0":
        deps += ["torchvision==0.19.0", "torchaudio==2.4.0"]
    elif v == "2.4.1":
        deps += ["torchvision==0.19.1", "torchaudio==2.4.1"]
    elif v == "2.5.0":
        deps += ["torchvision==0.20.0", "torchaudio==2.5.0"]
    else:
        set_github_warning(
            "UPDATE NEEDED",
            f"Leaving torchvision and torchaudio unpinned for pytorch=={v}",
        )
        deps += ["torchvision", "torchaudio"]

    # avoid `undefined symbol: iJIT_NotifyEvent` from `torch/lib/libtorch_cpu.so`
    # see https://github.com/pytorch/pytorch/issues/123097
    if (
        pytorch_version
        < Version(
            "2.1.0"  # TODO: check if this is the correct cutoff where the fix is not longer needed
        )
    ):
        deps.append("mkl ==2024.0.0")

    if pytorch_version < Version("2.2"):
        # avoid ImportError: cannot import name 'packaging' from 'pkg_resources'
        # see https://github.com/pypa/setuptools/issues/4376#issuecomment-2126162839
        deps.append("setuptools <70.0.0")

    if pytorch_version < Version(
        "2.3"
    ):  # TODO: verify that future pytorch 2.4 is numpy 2.0 compatible
        # make sure we have a compatible numpy version
        # see https://github.com/pytorch/vision/issues/8460
        deps.append("numpy <2")
    else:
        deps.append("numpy >=2,<3")

    return BioimageioCondaEnv(channels=channels, dependencies=deps)


def _get_default_onnx_env(*, opset_version: Optional[int]) -> BioimageioCondaEnv:
    if opset_version is None:
        opset_version = 15

    # note: we should not need to worry about the opset version,
    # see https://github.com/microsoft/onnxruntime/blob/master/docs/Versioning.md
    return BioimageioCondaEnv(dependencies=["onnxruntime"])


def _get_default_tf_env(tensorflow_version: Optional[Version]) -> BioimageioCondaEnv:
    if tensorflow_version is None or tensorflow_version.major < 2:
        tensorflow_version = Version("2.17")

    return BioimageioCondaEnv(
        dependencies=["bioimageio.core", f"tensorflow =={tensorflow_version}"],
    )


def _get_env_from_deps(
    deps: Union[v0_4.Dependencies, FileDescr],
) -> BioimageioCondaEnv:
    if isinstance(deps, v0_4.Dependencies):
        deps_reader = get_reader(deps.file)
        if deps.manager == "pip":
            pip_deps_str = deps_reader.read_text()
            pip_deps = [d.strip() for d in pip_deps_str.split("\n")]
            if "bioimageio.core" not in pip_deps:
                pip_deps.append("bioimageio.core")

            return BioimageioCondaEnv(
                dependencies=[PipDeps(pip=pip_deps)],
            )
        elif deps.manager in ("conda", "mamba"):
            return BioimageioCondaEnv.model_validate(read_yaml(deps_reader))
        else:
            raise ValueError(f"Dependency manager {deps.manager} not supported")

    elif isinstance(deps, FileDescr):
        deps_reader = deps.get_reader()
        return BioimageioCondaEnv.model_validate(read_yaml(deps_reader))
    else:
        assert_never(deps)



from typing import TYPE_CHECKING, ClassVar, Literal, Optional

from .._internal.common_nodes import Node
from .._internal.url import HttpUrl as HttpUrl
from ..generic.v0_2 import VALID_COVER_IMAGE_EXTENSIONS as VALID_COVER_IMAGE_EXTENSIONS
from ..generic.v0_2 import AttachmentsDescr as AttachmentsDescr
from ..generic.v0_2 import Author as Author
from ..generic.v0_2 import BadgeDescr as BadgeDescr
from ..generic.v0_2 import CiteEntry as CiteEntry
from ..generic.v0_2 import Doi as Doi
from ..generic.v0_2 import GenericDescrBase, ResourceId
from ..generic.v0_2 import LinkedResource as LinkedResource
from ..generic.v0_2 import Maintainer as Maintainer
from ..generic.v0_2 import OrcidId as OrcidId
from ..generic.v0_2 import RelativeFilePath as RelativeFilePath
from ..generic.v0_2 import Uploader as Uploader
from ..generic.v0_2 import Version as Version


class DatasetId(ResourceId):
    pass


class DatasetDescr(GenericDescrBase):
    """A bioimage.io dataset resource description file (dataset RDF) describes a dataset relevant to bioimage
    processing.
    """

    implemented_type: ClassVar[Literal["dataset"]] = "dataset"
    if TYPE_CHECKING:
        type: Literal["dataset"] = "dataset"
    else:
        type: Literal["dataset"]

    id: Optional[DatasetId] = None
    """bioimage.io-wide unique resource identifier
    assigned by bioimage.io; version **un**specific."""

    source: Optional[HttpUrl] = None
    """"URL to the source of the dataset."""


class LinkedDataset(Node):
    """Reference to a bioimage.io dataset."""

    id: DatasetId
    """A valid dataset `id` from the bioimage.io collection."""

    version_number: Optional[int] = None
    """version number (n-th published version, not the semantic version) of linked dataset"""



import os
import tempfile
import warnings
from contextlib import nullcontext
from functools import cache
from pathlib import Path
from typing import Optional, Union

from loguru import logger

from bioimageio.spec import save_bioimageio_package_as_folder
from bioimageio.spec._internal.validation_context import get_validation_context
from bioimageio.spec.model.v0_5 import ModelDescr

from ._hf_card import create_huggingface_model_card
from ._version import VERSION


@cache
def get_huggingface_api():  # pragma: no cover
    from huggingface_hub import HfApi

    return HfApi(library_name="bioimageio.spec", library_version=VERSION)


def push_to_hub(
    descr: ModelDescr,
    username_or_org: str,
    *,
    prep_dir: Optional[Union[os.PathLike[str], str]] = None,
    prep_only_no_upload: bool = False,
    create_pr: Optional[bool] = None,
):
    """Push the model package described by `descr` to the Hugging Face Hub.

    Note:
        - Uses `descr.id` as the repository name under the provided `username_or_org`.
        - If `descr.version` is set, the model package is uploaded to the 'main' branch
          and tagged with the version.
        - If `descr.version` is `None`, the model package is uploaded to the 'draft' branch.

    Args:
        descr: The model description to be pushed to the Hugging Face Hub.
        username_or_org: The Hugging Face username or organization under which the model package will be uploaded.
            The model ID from `descr.id` will be used as the repository name.
        prep_dir: Optional path to an empty directory where the model package will be prepared before uploading.
        prep_only_no_upload: If `True`, only prepare the model package in `prep_dir` without uploading it
            to the Hugging Face Hub.
        create_pr: If `False` commit directly to the 'main'/'draft' branch.
            If `True`, create a pull request targeting 'main'/'draft'.
            Defaults to `True` if uploading to a model description with version (to the main branch),
            and `False` if uploading a model description without version (to the 'draft' branch).

    Examples:
        Upload a model description as a new version to the main branch
        (`descr.id` and `descr.version` must be set):

        >>> descr = ModelDescr(id="my-model-id", version="1.0", create_pr=False, ...)  # doctest: +SKIP
        >>> push_to_hub(descr, "my_hf_username") # doctest: +SKIP

        Upload a model description as a draft to the 'draft' branch
        (`descr.id` must be set; `descr.version` must be `None`):

        >>> descr = ModelDescr(id="my-model-id", version=None, ...)  # doctest: +SKIP
        >>> push_to_hub(descr, "my_hf_username") # doctest: +SKIP

        See what would be uploaded without actually uploading:

        >>> push_to_hub(..., prep_dir="empty_local_folder", prep_only_no_upload=True) # doctest: +SKIP

    """

    if descr.id is None:
        raise ValueError("descr.id must be set to push to Hugging Face Hub.")
    repo_id = f"{username_or_org}/{descr.id}"

    if prep_dir is None:
        ctxt = tempfile.TemporaryDirectory(suffix="_" + repo_id.replace("/", "_"))
    elif Path(prep_dir).exists() and any(Path(prep_dir).iterdir()):
        raise ValueError("Provided `prep_dir` is not empty.")
        # TODO: implement resuming upload
        # prep_dir: If a non-empty folder is provided, it will be attempted to continue an interrupted upload.
        # logger.info(f"Continuing upload from {prep_dir}")
        # if prep_only_no_upload:
        #     raise ValueError("`prep_only_no_upload` is True but `prep_dir` is non-empty.")
    else:
        ctxt = nullcontext(prep_dir)

    with ctxt as pdir:
        _push_to_hub_impl(
            descr,
            repo_id=repo_id,
            prep_dir=Path(pdir),
            prep_only=prep_only_no_upload,
            create_pr=create_pr,
        )


def _push_to_hub_impl(
    descr: ModelDescr,
    *,
    repo_id: str,
    prep_dir: Path,
    prep_only: bool,
    create_pr: Optional[bool],
):
    readme, referenced_files = create_huggingface_model_card(descr, repo_id=repo_id)
    referenced_files_subfolders = {"images"}
    assert not (
        unexpected := [
            rf
            for rf in referenced_files
            if not any(rf.startswith(f"{sf}/") for sf in referenced_files_subfolders)
        ]
    ), f"unexpected folder of referenced files: {unexpected}"

    logger.info(f"Preparing model for upload at {prep_dir}.")
    prep_dir.mkdir(parents=True, exist_ok=True)
    _ = (prep_dir / "README.md").write_text(readme, encoding="utf-8")
    for img_name, img_data in referenced_files.items():
        image_path = prep_dir / img_name
        image_path.parent.mkdir(parents=True, exist_ok=True)
        _ = image_path.write_bytes(img_data)

    with get_validation_context().replace(file_name="bioimageio.yaml"):
        _ = save_bioimageio_package_as_folder(descr, output_path=prep_dir / "package")

    logger.info(f"Prepared model for upload at {prep_dir}")

    commit_message = f"Upload {descr.version or 'draft'} with bioimageio.spec {VERSION}"
    commit_description = (
        f"Version comment: {descr.version_comment}" if descr.version_comment else None
    )

    if not prep_only:  # pragma: no cover
        logger.info(f"Pushing model '{descr.id}' to Hugging Face Hub")

        api = get_huggingface_api()
        repo_url = api.create_repo(repo_id=repo_id, exist_ok=True, repo_type="model")
        logger.info(f"Created repository at {repo_url}")

        existing_refs = api.list_repo_refs(
            repo_id=repo_id, repo_type="model", include_pull_requests=True
        )
        has_draft_ref = False
        has_tag = False
        for ref in existing_refs.branches + existing_refs.tags:
            if ref.name == str(descr.version):
                has_tag = True
            if ref.name == "draft":
                has_draft_ref = True

        if descr.version is None:
            revision = "draft"
            if not has_draft_ref:
                api.create_branch(repo_id=repo_id, branch="draft", repo_type="model")
        else:
            revision = None

        if create_pr is None:
            # default to creating a PR if commiting to main branch,
            # commit directly to 'draft' branch
            create_pr = revision is None

        commit_info = api.upload_folder(
            repo_id=repo_id,
            revision=revision,
            folder_path=prep_dir,
            delete_patterns=[f"{sf}/*" for sf in referenced_files_subfolders]
            + ["package/*"],
            commit_message=commit_message,
            commit_description=commit_description,
            create_pr=create_pr,
        )
        logger.info(f"Created commit {commit_info.commit_url}")
        if descr.version is not None:
            if has_tag:
                warnings.warn(f"Moving existing version tag {descr.version}.")

            api.create_tag(
                repo_id=repo_id,
                tag=str(descr.version),
                revision=commit_info.oid,
                tag_message=descr.version_comment,
                exist_ok=True,
            )



import dataclasses
import inspect
from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Union, get_args

import pydantic.functional_validators
from annotated_types import BaseMetadata, GroupedMetadata
from loguru import logger
from pydantic import TypeAdapter
from pydantic_core import PydanticCustomError
from pydantic_core.core_schema import (
    NoInfoValidatorFunction,
    ValidationInfo,
    WithInfoValidatorFunction,
)
from typing_extensions import Annotated, LiteralString

from .utils import SLOTS
from .validation_context import get_validation_context
from .warning_levels import WARNING, WarningSeverity

if TYPE_CHECKING:
    from pydantic.functional_validators import _V2Validator  # type: ignore


ValidatorFunction = Union[NoInfoValidatorFunction, WithInfoValidatorFunction]

AnnotationMetaData = Union[BaseMetadata, GroupedMetadata]


def warn(
    typ: Union[AnnotationMetaData, Any],
    msg: LiteralString,  # warning message, e.g. "'{value}' incompatible with {typ}
    severity: WarningSeverity = WARNING,
):
    """treat a type or its annotation metadata as a warning condition"""
    if isinstance(typ, get_args(AnnotationMetaData)):
        typ = Annotated[Any, typ]

    validator: TypeAdapter[Any] = TypeAdapter(typ)

    return AfterWarner(
        validator.validate_python, severity=severity, msg=msg, context={"typ": typ}
    )


def _call_validator_func(
    func: "_V2Validator",
    mode: Literal["after", "before", "plain", "wrap"],
    value: Any,
    info: ValidationInfo,
) -> Any:
    # determine if validator needs info arg
    # logic adapted from pydantic._internal._decorators.py v2.11.10
    sig = inspect.signature(func)
    parameters = list(sig.parameters.values())
    n_positional = sum(
        1
        for param in parameters
        if param.kind
        in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
        # First argument is the value being validated/serialized, and can have a default value
        # (e.g. `float`, which has signature `(x=0, /)`). We assume other parameters (the info arg
        # for instance) should be required, and thus without any default value.
        and (param.default is inspect.Parameter.empty or param is parameters[0])
    )
    needs_info = None
    if mode == "wrap":
        if n_positional == 3:
            needs_info = True
        elif n_positional == 2:
            needs_info = False
    else:
        assert mode in {"before", "after", "plain"}, (
            f"invalid mode: {mode!r}, expected 'before', 'after' or 'plain"
        )
        if n_positional == 2:
            needs_info = True
        elif n_positional == 1:
            needs_info = False

    assert needs_info is not None, "could not determine if validator needs info arg"
    if needs_info:
        return func(value, info)  # type: ignore
    else:
        return func(value)  # type: ignore


def as_warning(
    func: "_V2Validator",
    *,
    mode: Literal["after", "before", "plain", "wrap"] = "after",
    severity: WarningSeverity = WARNING,
    msg: Optional[LiteralString] = None,
    msg_context: Optional[Dict[str, Any]] = None,
) -> ValidatorFunction:
    """turn validation function into a no-op, based on warning level"""

    def wrapper(value: Any, info: ValidationInfo) -> Any:
        try:
            _call_validator_func(func, mode, value, info)
        except (AssertionError, ValueError) as e:
            issue_warning(
                msg or ",".join(e.args),
                field=info.field_name,
                log_depth=1,
                msg_context=msg_context,
                severity=severity,
                value=value,
            )

        return value

    return wrapper


@dataclasses.dataclass(frozen=True, **SLOTS)
class AfterWarner(pydantic.functional_validators.AfterValidator):
    """Like AfterValidator, but wraps validation `func` `as_warning`"""

    severity: WarningSeverity = WARNING
    msg: Optional[LiteralString] = None
    context: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        object.__setattr__(
            self,
            "func",
            as_warning(
                self.func,
                mode="after",
                severity=self.severity,
                msg=self.msg,
                msg_context=self.context,
            ),
        )


@dataclasses.dataclass(frozen=True, **SLOTS)
class BeforeWarner(pydantic.functional_validators.BeforeValidator):
    """Like BeforeValidator, but wraps validation `func` `as_warning`"""

    severity: WarningSeverity = WARNING
    msg: Optional[LiteralString] = None
    context: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        object.__setattr__(
            self,
            "func",
            as_warning(
                self.func,
                mode="before",
                severity=self.severity,
                msg=self.msg,
                msg_context=self.context,
            ),
        )


# TODO: add `loc: Loc` to `issue_warning()`
#   and use a loguru handler to format warnings accordingly
def issue_warning(
    msg: LiteralString,
    *,
    value: Any,
    severity: WarningSeverity = WARNING,
    msg_context: Optional[Dict[str, Any]] = None,
    field: Optional[str] = None,
    log_depth: int = 1,
):
    msg_context = {"value": value, "severity": severity, **(msg_context or {})}

    if severity >= (ctxt := get_validation_context()).warning_level:
        raise PydanticCustomError("warning", msg, msg_context)
    elif ctxt.log_warnings:
        log_msg = (field + ": " if field else "") + (msg.format(**msg_context))
        logger.opt(depth=log_depth).log(severity, log_msg)



from __future__ import annotations

from typing import Any, Optional, Tuple, Union

import packaging.version
from pydantic import PrivateAttr, RootModel


class Version(RootModel[Union[str, int, float]]):
    """wraps a packaging.version.Version instance for validation in pydantic models"""

    _version: packaging.version.Version = PrivateAttr()

    def __str__(self):
        return str(self._version)

    def model_post_init(self, __context: Any) -> None:
        """set `_version` attribute @private"""
        self._version = packaging.version.Version(str(self.root))
        return super().model_post_init(__context)

    def __lt__(self, other: Version):
        return self._version < other._version

    def __eq__(self, other: Version):
        return self._version == other._version

    # the properties below are adopted from and mirror properties of packaging.version.Version
    @property
    def epoch(self) -> int:
        """The epoch of the version.

        >>> Version("2.0.0").epoch
        0
        >>> Version("1!2.0.0").epoch
        1
        """
        return self._version.epoch

    @property
    def release(self) -> Tuple[int, ...]:
        """The components of the "release" segment of the version.

        >>> Version("1.2.3").release
        (1, 2, 3)
        >>> Version("2.0.0").release
        (2, 0, 0)
        >>> Version("1!2.0.0.post0").release
        (2, 0, 0)

        Includes trailing zeroes but not the epoch or any pre-release / development /
        post-release suffixes.
        """
        return self._version.release

    @property
    def pre(self) -> Optional[Tuple[str, int]]:
        """The pre-release segment of the version.

        >>> print(Version("1.2.3").pre)
        None
        >>> Version("1.2.3a1").pre
        ('a', 1)
        >>> Version("1.2.3b1").pre
        ('b', 1)
        >>> Version("1.2.3rc1").pre
        ('rc', 1)
        """
        return self._version.pre

    @property
    def post(self) -> Optional[int]:
        """The post-release number of the version.

        >>> print(Version("1.2.3").post)
        None
        >>> Version("1.2.3.post1").post
        1
        """
        return self._version.post

    @property
    def dev(self) -> Optional[int]:
        """The development number of the version.

        >>> print(Version("1.2.3").dev)
        None
        >>> Version("1.2.3.dev1").dev
        1
        """
        return self._version.dev

    @property
    def local(self) -> Optional[str]:
        """The local version segment of the version.

        >>> print(Version("1.2.3").local)
        None
        >>> Version("1.2.3+abc").local
        'abc'
        """
        return self._version.local

    @property
    def public(self) -> str:
        """The public portion of the version.

        >>> Version("1.2.3").public
        '1.2.3'
        >>> Version("1.2.3+abc").public
        '1.2.3'
        >>> Version("1.2.3+abc.dev1").public
        '1.2.3'
        """
        return self._version.public

    @property
    def base_version(self) -> str:
        """The "base version" of the version.

        >>> Version("1.2.3").base_version
        '1.2.3'
        >>> Version("1.2.3+abc").base_version
        '1.2.3'
        >>> Version("1!1.2.3+abc.dev1").base_version
        '1!1.2.3'

        The "base version" is the public version of the project without any pre or post
        release markers.
        """
        return self._version.base_version

    @property
    def is_prerelease(self) -> bool:
        """Whether this version is a pre-release.

        >>> Version("1.2.3").is_prerelease
        False
        >>> Version("1.2.3a1").is_prerelease
        True
        >>> Version("1.2.3b1").is_prerelease
        True
        >>> Version("1.2.3rc1").is_prerelease
        True
        >>> Version("1.2.3dev1").is_prerelease
        True
        """
        return self._version.is_prerelease

    @property
    def is_postrelease(self) -> bool:
        """Whether this version is a post-release.

        >>> Version("1.2.3").is_postrelease
        False
        >>> Version("1.2.3.post1").is_postrelease
        True
        """
        return self._version.is_postrelease

    @property
    def is_devrelease(self) -> bool:
        """Whether this version is a development release.

        >>> Version("1.2.3").is_devrelease
        False
        >>> Version("1.2.3.dev1").is_devrelease
        True
        """
        return self._version.is_devrelease

    @property
    def major(self) -> int:
        """The first item of :attr:`release` or ``0`` if unavailable.

        >>> Version("1.2.3").major
        1
        """
        return self._version.major

    @property
    def minor(self) -> int:
        """The second item of :attr:`release` or ``0`` if unavailable.

        >>> Version("1.2.3").minor
        2
        >>> Version("1").minor
        0
        """
        return self._version.minor

    @property
    def micro(self) -> int:
        """The third item of :attr:`release` or ``0`` if unavailable.

        >>> Version("1.2.3").micro
        3
        >>> Version("1").micro
        0
        """
        return self._version.micro



# autogen: start
"""application resource description format

Implementations of all released minor versions are available in submodules.
"""

from typing import Union

from pydantic import Discriminator, Field
from typing_extensions import Annotated

from . import v0_2, v0_3

ApplicationDescr = v0_3.ApplicationDescr
ApplicationDescr_v0_2 = v0_2.ApplicationDescr
ApplicationDescr_v0_3 = v0_3.ApplicationDescr

AnyApplicationDescr = Annotated[
    Union[
        Annotated[ApplicationDescr_v0_2, Field(title="application 0.2")],
        Annotated[ApplicationDescr_v0_3, Field(title="application 0.3")],
    ],
    Discriminator("format_version"),
    Field(title="application"),
]
"""Union of any released application desription"""
# autogen: stop



VERSION = "0.5.7.1"
"""bioimageio.spec version as MAJOR.MINOR.PATCH.LIB

MAJOR.MINOR.PATCH correspond to the latest model description format version implemented.

LIB may be bumped for library changes that do not affect the format version.
"""



import os
from functools import cached_property
from pathlib import Path
from typing import Optional, Union

import platformdirs
from genericache import DiskCache
from genericache.digest import UrlDigest
from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Annotated

from .root_url import RootHttpUrl


class Settings(
    BaseSettings, extra="ignore", allow_inf_nan=False, validate_assignment=True
):
    """environment variables for bioimageio.spec"""

    model_config = SettingsConfigDict(
        env_prefix="BIOIMAGEIO_", env_file=".env", env_file_encoding="utf-8"
    )

    allow_pickle: bool = False
    """Sets the `allow_pickle` argument for `numpy.load()`"""

    cache_path: Path = Path(platformdirs.user_cache_dir("bioimageio"))
    """bioimageio cache location"""

    @field_validator("cache_path", mode="after")
    @classmethod
    def _expand_user(cls, value: Path):
        return Path(os.path.expanduser(str(value)))

    CI: Annotated[Union[bool, str], Field(alias="CI")] = False
    """Wether or not the execution happens in a continuous integration (CI) environment."""

    collection_http_pattern: str = (
        "https://hypha.aicell.io/bioimage-io/artifacts/{bioimageio_id}/files/rdf.yaml"
    )
    """A pattern to map bioimageio IDs to bioimageio.yaml URLs.
    Notes:
    - '{bioimageio_id}' is replaced with user query,
      e.g. "affable-shark" when calling `load_description("affable-shark")`.
    - This method takes precedence over resolving via `id_map`.
    - If this endpoints fails, we fall back to `id_map`.
    """

    github_username: Optional[str] = None
    """GitHub username for API requests"""

    github_token: Optional[str] = None
    """GitHub token for API requests"""

    http_timeout: float = 10.0
    """Timeout in seconds for http requests."""

    huggingface_http_pattern: str = (
        "https://huggingface.co/{repo_id}/tree/{branch}/package/bioimageio.yaml"
    )
    """A pattern to map huggingface repo IDs to bioimageio.yaml URLs.
    Notes:
    - Used for loading source strings of the form "huggingface/{user_or_org}/{resource_id}[/{version}]"
    - example use: `load_description("huggingface/fynnbe/ambitious-sloth/1.3")`
    - A given version {version} is mapped to a branch name "v{version}", e.g. "v1.3".
    - If no version is provided the "main" branch is used.
    - This method takes precedence over resolving via `id_map`.
    - If this endpoints fails, we fall back to `id_map`.
    """

    hypha_upload: str = (
        "https://hypha.aicell.io/public/services/artifact-manager/create"
    )
    """URL to the upload endpoint for bioimageio resources."""

    hypha_upload_token: Optional[str] = None
    """Hypha API token to use for uploads.

    By setting this token you agree to our terms of service at https://bioimage.io/#/toc.

    How to obtain a token:
        1. Login to https://bioimage.io
        2. Generate a new token at https://bioimage.io/#/api?tab=hypha-rpc
    """

    id_map: str = (
        "https://uk1s3.embassy.ebi.ac.uk/public-datasets/bioimage.io/id_map.json"
    )
    """URL to bioimageio id_map.json to resolve resource IDs."""

    id_map_draft: str = (
        "https://uk1s3.embassy.ebi.ac.uk/public-datasets/bioimage.io/id_map_draft.json"
    )
    """URL to bioimageio id_map_draft.json to resolve draft IDs ending with '/draft'."""

    log_warnings: bool = True
    """Log validation warnings to console."""

    perform_io_checks: bool = True
    """Wether or not to perform validation that requires file io,
    e.g. downloading a remote files.

    Existence of any local absolute file paths is still being checked."""

    resolve_draft: bool = True
    """Flag to resolve draft resource versions following the pattern
    <resource id>/draft.

    Note that anyone may stage a new draft and that such a draft version
    may not have been reviewed yet.
    Set this flag to False to avoid this potential security risk
    and disallow loading draft versions."""

    user_agent: Optional[str] = None
    """user agent for http requests"""

    @cached_property
    def disk_cache(self):
        cache = DiskCache[RootHttpUrl].create(
            url_type=RootHttpUrl,
            cache_dir=self.cache_path,
            url_hasher=UrlDigest.from_str,
        )
        return cache

    @property
    def github_auth(self):
        if self.github_username is None or self.github_token is None:
            return None
        else:
            return (self.github_username, self.github_token)


settings = Settings()
"""parsed environment variables for bioimageio.spec"""



from pathlib import Path

import pytest

from bioimageio.spec._internal.io import BioimageioYamlContent
from bioimageio.spec._internal.validation_context import ValidationContext


def test_load_non_existing_rdf():
    from bioimageio.spec import load_description

    spec_path = Path("some/none/existing/path/to/spec.model.yaml")

    with pytest.raises(FileNotFoundError):
        _ = load_description(spec_path)


@pytest.mark.parametrize(
    "rid",
    [
        "invigorating-lab-coat",
        "invigorating-lab-coat/1",
        "10.5281/zenodo.11092896",  # backup doi of version 1
        "10.5281/zenodo.11092895",  # concept doi of backup
    ],
)
def test_load_by_id(rid: str):
    from bioimageio.spec._internal.io_utils import open_bioimageio_yaml

    rdf = open_bioimageio_yaml(rid).content
    assert rdf["id"] == "invigorating-lab-coat"


def test_load_description_again(unet2d_data: BioimageioYamlContent):
    from bioimageio.spec import build_description, load_description

    descr = build_description(
        unet2d_data, context=ValidationContext(perform_io_checks=False)
    )
    descr2 = load_description(
        descr,  # pyright: ignore[reportArgumentType]
        perform_io_checks=False,
    )
    assert descr is descr2


def test_load_dataset_description(covid_if_dataset_path: Path, tmp_path: Path):
    from bioimageio.spec import load_dataset_description
    from bioimageio.spec._io import save_bioimageio_yaml_only
    from bioimageio.spec.dataset.v0_2 import DatasetDescr

    dataset_descr = load_dataset_description(covid_if_dataset_path)
    assert isinstance(dataset_descr, DatasetDescr)

    # this example happens to consist only of the bioimageio.yaml file,
    # so we can test the roundtrip with `save_bioimageio_yaml_only`
    save_bioimageio_yaml_only(dataset_descr, tmp_path / "dataset.yaml")
    dataset_descr2 = load_dataset_description(tmp_path / "dataset.yaml")
    assert isinstance(dataset_descr2, DatasetDescr)  # we cannot expect
    assert dataset_descr.model_dump() == dataset_descr2.model_dump()



# pragma: no cover
# type: ignore
"""ImJoy plugin parser module."""

import copy
import json
import uuid
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, Tuple, Union
from urllib.parse import urljoin

import httpx
from pydantic import DirectoryPath, FilePath, HttpUrl
from ruyaml import YAML

yaml = YAML(typ="safe")

tag_types = ["config", "script", "link", "window", "style", "docs", "attachment"]

CONFIGURABLE_FIELDS = [
    "env",
    "requirements",
    "dependencies",
    "icon",
    "ui",
    "type",
    "flags",
    "labels",
    "cover",
    "base_frame",
    "base_worker",
    "passive",
]


class dotdict(dict):  # pylint: disable=invalid-name
    """Access dictionary attributes with dot.notation."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __deepcopy__(self, memo=None):
        """Make a deep copy."""
        return dotdict(copy.deepcopy(dict(self), memo=memo))


def parse_imjoy_plugin(source, overwrite_config=None):
    """Parse ImJoy plugin file and return a dict with all the fields."""
    from lxml import etree

    root = etree.HTML("<html>" + source + "</html>")
    plugin_comp = dotdict()
    for tag_type in tag_types:
        elms = root.xpath(f".//{tag_type}")
        values = []
        for elm in elms:
            values.append(
                dotdict(
                    attrs=dotdict(elm.attrib),
                    content=elm.text,
                )
            )
        plugin_comp[tag_type] = values
    if plugin_comp.config[0].attrs.lang == "yaml":
        config = yaml.load(plugin_comp.config[0].content)
    elif plugin_comp.config[0].attrs.lang == "json":
        config = json.loads(plugin_comp.config[0].content)
    else:
        raise Exception(
            "Unsupported config language: " + plugin_comp.config[0].attrs.lang
        )

    overwrite_config = overwrite_config or {}
    config["tag"] = overwrite_config.get("tag") or (
        config.get("tags") and config.get("tags")[0]
    )
    config["hot_reloading"] = overwrite_config.get("hot_reloading")
    config["scripts"] = []
    # try to match the script with current tag
    for elm in plugin_comp.script:
        if elm.attrs.tag == config["tag"]:
            config["script"] = elm.content
        # exclude script with mismatched tag
        if not elm.attrs.tag or elm.attrs.tag == config["tag"]:
            config["scripts"].append(elm)
    if not config.get("script") and len(plugin_comp.script) > 0:
        config["script"] = plugin_comp.script[0].content
        config["lang"] = plugin_comp.script[0].attrs.lang
    config["links"] = plugin_comp.link or None
    config["windows"] = plugin_comp.window or None
    config["styles"] = plugin_comp.style or None
    config["docs"] = plugin_comp.docs[0] if plugin_comp.docs else config.get("docs")
    config["attachments"] = plugin_comp.attachment or None

    config["_id"] = overwrite_config.get("_id") or config.get("name").replace(" ", "_")
    config["uri"] = overwrite_config.get("uri")
    config["origin"] = overwrite_config.get("origin")
    config["namespace"] = overwrite_config.get("namespace")
    config["code"] = source
    config["id"] = (
        config.get("name").strip().replace(" ", "_") + "_" + str(uuid.uuid4())
    )
    config["runnable"] = config.get("runnable", True)
    config["requirements"] = config.get("requirements") or []

    for field in CONFIGURABLE_FIELDS:
        obj = config.get(field)
        if obj and isinstance(obj, dict) and not isinstance(obj, list):
            if config.get("tag"):
                config[field] = obj.get(config.get("tag"))
                if not obj.get(config.get("tag")):
                    print(
                        "WARNING: "
                        + field
                        + " do not contain a tag named: "
                        + config.get("tag")
                    )
            else:
                raise Exception("You must use 'tags' with configurable fields.")
    config["lang"] = config.get("lang") or "javascript"
    return config


def convert_config_to_rdf(plugin_config, source_url=None) -> dict:
    """Convert imjoy plugin config to RDF format."""
    rdf = {
        "type": "application",
    }
    if source_url:
        rdf["source"] = source_url
    fields = [
        "icon",
        "name",
        "version",
        "api_version",
        "description",
        "license",
        "requirements",
        "dependencies",
        "env",
        "passive",
        "services",
    ]
    for field in fields:
        if field in plugin_config:
            rdf[field] = plugin_config[field]
    tags = plugin_config.get("labels", []) + plugin_config.get("flags", [])
    if "bioengine" not in tags:
        tags.append("bioengine")
    rdf["tags"] = tags

    # docs = plugin_config.get("docs")
    #     if isinstance(docs, dict):
    #         rdf["documentation"] = docs.get("content")
    #     elif isinstance(docs, str):
    #         rdf["documentation"] = docs
    rdf["covers"] = plugin_config.get("cover")
    # make sure we have a list
    if not rdf["covers"]:
        rdf["covers"] = []
    elif not isinstance(rdf["covers"], list):
        rdf["covers"] = [rdf["covers"]]

    rdf["badges"] = plugin_config.get("badge")
    if not rdf["badges"]:
        rdf["badges"] = []
    elif not isinstance(rdf["badges"], list):
        rdf["badges"] = [rdf["badges"]]

    authors = plugin_config.get("author")
    if authors:
        if isinstance(authors, str):
            authors = {"name": authors}
        if not isinstance(authors, list):
            authors = [authors]
        rdf["authors"] = authors

    return rdf


def get_plugin_as_rdf(source_url: str) -> Dict[Any, Any]:
    """Get imjoy plugin config in RDF format."""
    req = httpx.get(source_url, timeout=5)
    source = req.text
    plugin_config = parse_imjoy_plugin(source)
    rdf = convert_config_to_rdf(plugin_config, source_url)
    return rdf


def enrich_partial_rdf_with_imjoy_plugin(
    partial_rdf: Dict[str, Any],
    root: Union[HttpUrl, DirectoryPath],
    resolve_rdf_source: Callable[
        [Union[HttpUrl, FilePath, str]],
        Tuple[Dict[str, Any], str, Union[HttpUrl, DirectoryPath]],
    ],
) -> Dict[str, Any]:
    """
    a (partial) rdf may have 'rdf_source' or 'source' which resolve to rdf data that may be overwritten.
    """

    enriched_rdf: Dict[str, Any] = {}
    if "rdf_source" in partial_rdf:
        given_rdf_src = partial_rdf["rdf_source"]
        if isinstance(given_rdf_src, str) and given_rdf_src.split("?")[0].endswith(
            ".imjoy.html"
        ):
            # given_rdf_src is an imjoy plugin
            rdf_source = dict(get_plugin_as_rdf(given_rdf_src))
        else:
            # given_rdf_src is an actual rdf
            if isinstance(given_rdf_src, dict):
                rdf_source: Dict[str, Any] = given_rdf_src
            else:
                try:
                    rdf_source, _, rdf_source_root = resolve_rdf_source(given_rdf_src)
                except Exception as e:
                    try:
                        rdf_source, _, rdf_source_root = resolve_rdf_source(
                            root / given_rdf_src
                            if isinstance(root, Path)
                            else urljoin(str(root), given_rdf_src)
                        )
                    except Exception as ee:
                        rdf_source = {}
                        warnings.warn(
                            f"Failed to resolve `rdf_source`: 1. {e}\n2. {ee}"
                        )
                    else:
                        rdf_source["root_path"] = (
                            rdf_source_root  # enables remote source content to be resolved
                        )
                else:
                    rdf_source["root_path"] = (
                        rdf_source_root  # enables remote source content to be resolved
                    )

        assert isinstance(rdf_source, dict)
        enriched_rdf.update(rdf_source)

    if "source" in partial_rdf:
        if partial_rdf["source"].split("?")[0].endswith(".imjoy.html"):
            rdf_from_source = get_plugin_as_rdf(partial_rdf["source"])
            enriched_rdf.update(rdf_from_source)

    enriched_rdf.update(
        partial_rdf
    )  # initial partial rdf overwrites fields from rdf_source or source
    return enriched_rdf



"""bioimageio.spec --- BioImage.IO specifications in Python.

This package provides Pydantic data models for BioImage.IO resource descriptions and tools to work with such descriptions.

The BioImage.IO resource description format is resource type specific (e.g. models, datasets) and versioned.
bioimageio.spec defines and validates these specifications and is backwards compatible with previous format versions.
It also provides download/upload to/from the BioImage.IO Model Zoo at https://bioimage.io.

Note:
    For additional tools to work with BioImage.IO resources in Python, consider using the [bioimageio.core](https://bioimage-io.github.io/core-bioimage-io-python) package.
"""

# ruff: noqa: E402
from loguru import logger

logger.disable("bioimageio.spec")

from . import _version
from . import application as application
from . import common as common
from . import conda_env as conda_env
from . import dataset as dataset
from . import generic as generic
from . import model as model
from . import notebook as notebook
from . import summary as summary
from . import utils as utils
from ._description import LatestResourceDescr as LatestResourceDescr
from ._description import ResourceDescr as ResourceDescr
from ._description import SpecificResourceDescr as SpecificResourceDescr
from ._description import build_description as build_description
from ._description import dump_description as dump_description
from ._description import validate_format as validate_format
from ._get_conda_env import get_conda_env as get_conda_env
from ._internal import common_nodes, validation_context
from ._internal import settings as settings
from ._io import load_dataset_description as load_dataset_description
from ._io import load_description as load_description
from ._io import (
    load_description_and_validate_format_only as load_description_and_validate_format_only,
)
from ._io import load_model_description as load_model_description
from ._io import save_bioimageio_yaml_only as save_bioimageio_yaml_only
from ._io import update_format as update_format
from ._io import update_hashes as update_hashes
from ._package import get_resource_package_content as get_resource_package_content
from ._package import save_bioimageio_package as save_bioimageio_package
from ._package import (
    save_bioimageio_package_as_folder as save_bioimageio_package_as_folder,
)
from ._package import (
    save_bioimageio_package_to_stream as save_bioimageio_package_to_stream,
)
from ._pretty_validation_errors import (
    PRETTY_VALIDATION_ERRORS_IN_IPYNB_ENABLED as PRETTY_VALIDATION_ERRORS_IN_IPYNB_ENABLED,
)
from ._upload import upload as upload

__version__ = _version.VERSION

# reexpose slected objects from submodules
AnyApplicationDescr = application.AnyApplicationDescr
AnyDatasetDescr = dataset.AnyDatasetDescr
AnyGenericDescr = generic.AnyGenericDescr
AnyModelDescr = model.AnyModelDescr
AnyNotebookDescr = notebook.AnyNotebookDescr
ApplicationDescr = application.ApplicationDescr
BioimageioCondaEnv = conda_env.BioimageioCondaEnv
DatasetDescr = dataset.DatasetDescr
GenericDescr = generic.GenericDescr
get_validation_context = validation_context.get_validation_context
InvalidDescr = common_nodes.InvalidDescr
ModelDescr = model.ModelDescr
NotebookDescr = notebook.NotebookDescr
ValidationContext = validation_context.ValidationContext
ValidationSummary = summary.ValidationSummary



from contextlib import nullcontext
from pathlib import PurePosixPath
from typing import Any, ClassVar, Optional, Type, Union

import httpx
import pydantic
from loguru import logger
from pydantic import RootModel
from typing_extensions import Literal, assert_never

from . import warning_levels
from ._settings import settings
from .field_warning import issue_warning
from .root_url import RootHttpUrl
from .validation_context import get_validation_context


def _validate_url(url: Union[str, pydantic.HttpUrl]) -> pydantic.HttpUrl:
    return _validate_url_impl(url, request_mode="head", timeout=settings.http_timeout)


def _validate_url_impl(
    url: Union[str, pydantic.HttpUrl],
    request_mode: Literal["head", "get_stream", "get"],
    timeout: float,
) -> pydantic.HttpUrl:
    url = str(url)
    context = get_validation_context()
    if url in context.known_files:
        return pydantic.HttpUrl(url)

    val_url = url

    if url.startswith("http://example.com") or url.startswith("https://example.com"):
        return pydantic.HttpUrl(url)

    if url.startswith("https://colab.research.google.com/github/"):
        # get requests for colab returns 200 even if the source notebook does not exists.
        # We therefore validate the url to the notebbok instead (for github notebooks)
        val_url = url.replace(
            "https://colab.research.google.com/github/", "https://github.com/"
        )
    elif url.startswith("https://colab.research.google.com/"):
        # TODO: improve validation of non-github colab urls
        issue_warning(
            "colab urls currently pass even if the notebook url was not found. Cannot fully validate {value}",
            value=url,
            severity=warning_levels.INFO,
        )

    try:
        if request_mode in ("head", "get"):
            request_ctxt = nullcontext(
                httpx.request(
                    request_mode.upper(),
                    val_url,
                    timeout=timeout,
                    follow_redirects=True,
                )
            )
        elif request_mode == "get_stream":
            request_ctxt = httpx.stream(
                "GET", val_url, timeout=timeout, follow_redirects=True
            )
        else:
            assert_never(request_mode)

        with request_ctxt as r:
            status_code = r.status_code
            reason = r.reason_phrase
            location = r.headers.get("location")

    except (
        httpx.InvalidURL,
        httpx.TooManyRedirects,
    ) as e:
        raise ValueError(f"Invalid URL '{url}': {e}")
    except httpx.RequestError as e:
        issue_warning(
            "Failed to validate URL '{value}': {error}\nrequest: {request}",
            value=url,
            msg_context={"error": str(e), "request": e.request},
        )
    except Exception as e:
        issue_warning(
            "Failed to validate URL '{value}': {error}",
            value=url,
            msg_context={"error": str(e)},
        )
    else:
        if status_code == 200:  # ok
            pass
        elif status_code in (302, 303):  # found
            pass
        elif status_code in (301, 308):
            issue_warning(
                "URL redirected ({status_code}): consider updating {value} with new"
                + " location: {location}",
                value=url,
                severity=warning_levels.INFO,
                msg_context={
                    "status_code": status_code,
                    "location": location,
                },
            )
        elif request_mode == "head":
            return _validate_url_impl(url, request_mode="get_stream", timeout=timeout)
        elif request_mode == "get_stream":
            return _validate_url_impl(url, request_mode="get", timeout=timeout)
        elif request_mode == "get":
            issue_warning(
                "{status_code}: {reason} ({value})",
                value=url,
                severity=(
                    warning_levels.INFO
                    if status_code == 405  # may be returned due to a captcha
                    else warning_levels.WARNING
                ),
                msg_context={
                    "status_code": status_code,
                    "reason": reason,
                },
            )
        else:
            assert_never(request_mode)

    context.known_files[url] = None
    return pydantic.HttpUrl(url)


class HttpUrl(RootHttpUrl):
    """A URL with the HTTP or HTTPS scheme."""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[pydantic.HttpUrl]
    _exists: Optional[bool] = None

    def _after_validator(self):
        self = super()._after_validator()
        context = get_validation_context()
        if context.perform_io_checks:
            _ = self.exists()

        return self

    def exists(self):
        """True if URL is available"""
        if self._exists is None:
            ctxt = get_validation_context()
            try:
                with ctxt.replace(warning_level=warning_levels.WARNING):
                    self._validated = _validate_url(self._validated)
            except Exception as e:
                if ctxt.log_warnings:
                    logger.info(e)

                self._exists = False
            else:
                self._exists = True

        return self._exists

    @property
    def suffix(self) -> str:
        if self.path is None:
            return ""
        else:
            return PurePosixPath(self.path).suffix



# autogen: start
"""generic resource description format

Implementations of all released minor versions are available in submodules.
"""

from typing import Union

from pydantic import Discriminator, Field
from typing_extensions import Annotated

from . import v0_2, v0_3

GenericDescr = v0_3.GenericDescr
GenericDescr_v0_2 = v0_2.GenericDescr
GenericDescr_v0_3 = v0_3.GenericDescr

AnyGenericDescr = Annotated[
    Union[
        Annotated[GenericDescr_v0_2, Field(title="generic 0.2")],
        Annotated[GenericDescr_v0_3, Field(title="generic 0.3")],
    ],
    Discriminator("format_version"),
    Field(title="generic"),
]
"""Union of any released generic desription"""
# autogen: stop



from __future__ import annotations

from datetime import datetime, timezone
from keyword import iskeyword
from typing import Any, ClassVar, Sequence, Type, TypeVar, Union

import annotated_types
from dateutil.parser import isoparse
from pydantic import RootModel, StringConstraints
from typing_extensions import Annotated, Literal

from .constants import DOI_REGEX, SI_UNIT_REGEX
from .field_warning import AfterWarner
from .io import FileSource, PermissiveFileSource, RelativeFilePath
from .io_basics import AbsoluteDirectory, AbsoluteFilePath, FileName, FilePath, Sha256
from .io_packaging import FileSource_
from .license_id import DeprecatedLicenseId, LicenseId
from .type_guards import is_sequence
from .url import HttpUrl
from .utils import PrettyPlainSerializer
from .validated_string import ValidatedString
from .validator_annotations import AfterValidator, BeforeValidator
from .version_type import Version
from .warning_levels import ALERT

UTC = timezone.utc

__all__ = [
    "AbsoluteDirectory",
    "AbsoluteFilePath",
    "Datetime",
    "DeprecatedLicenseId",
    "Doi",
    "FileName",
    "FilePath",
    "FileSource",
    "FileSource_",
    "ImportantFileSource",
    "HttpUrl",
    "Identifier",
    "LicenseId",
    "LowerCaseIdentifier",
    "NotEmpty",
    "OrcidId",
    "PermissiveFileSource",
    "RelativeFilePath",
    "Sha256",
    "SiUnit",
    "Version",
]
S = TypeVar("S", bound=Sequence[Any])
A = TypeVar("A", bound=Any)
NotEmpty = Annotated[S, annotated_types.MinLen(1)]


def _validate_fair(value: Any) -> Any:
    """Raise trivial values."""
    if value is None or (is_sequence(value) and not value):
        raise ValueError("Needs to be filled for FAIR compliance")

    return value


FAIR = Annotated[
    A,
    AfterWarner(_validate_fair, severity=ALERT),
]

ImportantFileSource = FileSource_
"""DEPRECATED alias, use `FileSource` instead"""


def _validate_identifier(s: str) -> str:
    if not s.isidentifier():
        raise ValueError(
            f"'{s}' is not a valid (Python) identifier, see"
            + " https://docs.python.org/3/reference/lexical_analysis.html#identifiers"
            + " for details."
        )

    return s


def _validate_is_not_keyword(s: str) -> str:
    if iskeyword(s):
        raise ValueError(f"'{s}' is a Python keyword and not allowed here.")

    return s


def _validate_datetime(dt: Union[datetime, str, Any]) -> datetime:
    if isinstance(dt, datetime):
        return dt
    elif isinstance(dt, str):
        return isoparse(dt).astimezone(UTC)

    raise ValueError(f"'{dt}' not a string or datetime.")


def _validate_orcid_id(orcid_id: str):
    if len(orcid_id) == 19 and all(orcid_id[idx] == "-" for idx in [4, 9, 14]):
        check = 0
        for n in orcid_id[:4] + orcid_id[5:9] + orcid_id[10:14] + orcid_id[15:]:
            # adapted from stdnum.iso7064.mod_11_2.checksum()
            check = (2 * check + int(10 if n == "X" else n)) % 11
        if check == 1:
            return orcid_id  # valid

    raise ValueError(
        f"'{orcid_id} is not a valid ORCID iD in hyphenated groups of 4 digits."
    )


# TODO follow up on https://github.com/pydantic/pydantic/issues/8964
# to remove _serialize_datetime
def _serialize_datetime_json(dt: datetime) -> str:
    return dt.astimezone(UTC).isoformat(timespec="seconds")


class Datetime(
    RootModel[
        Annotated[
            datetime,
            BeforeValidator(_validate_datetime),
            PrettyPlainSerializer(
                _serialize_datetime_json, when_used="json-unless-none"
            ),
        ]
    ]
):
    """Timestamp in [ISO 8601](#https://en.wikipedia.org/wiki/ISO_8601) format
    with a few restrictions listed [here](https://docs.python.org/3/library/datetime.html#datetime.datetime.fromisoformat).
    """

    @classmethod
    def now(cls):
        return cls(datetime.now(UTC))


class Doi(ValidatedString):
    """A digital object identifier, see https://www.doi.org/"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[str, StringConstraints(pattern=DOI_REGEX)]
    ]


FormatVersionPlaceholder = Literal["latest", "discover"]
IdentifierAnno = Annotated[
    NotEmpty[str],
    AfterValidator(_validate_identifier),
    AfterValidator(_validate_is_not_keyword),
]


class Identifier(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[IdentifierAnno]


LowerCaseIdentifierAnno = Annotated[IdentifierAnno, annotated_types.LowerCase]


class LowerCaseIdentifier(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[LowerCaseIdentifierAnno]


class OrcidId(ValidatedString):
    """An ORCID identifier, see https://orcid.org/"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[str, AfterValidator(_validate_orcid_id)]
    ]


def _normalize_multiplication(si_unit: Union[Any, str]) -> Union[Any, str]:
    if isinstance(si_unit, str):
        return si_unit.replace("×", "·").replace("*", "·").replace(" ", "·")
    else:
        return si_unit


class SiUnit(ValidatedString):
    """An SI unit"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(min_length=1, pattern=SI_UNIT_REGEX),
            BeforeValidator(_normalize_multiplication),
        ]
    ]


RelativeTolerance = Annotated[float, annotated_types.Interval(ge=0, le=1e-2)]
AbsoluteTolerance = Annotated[float, annotated_types.Interval(ge=0)]
MismatchedElementsPerMillion = Annotated[int, annotated_types.Interval(ge=0, le=1000)]



import collections.abc
import string
from pathlib import Path

import imageio
from loguru import logger

from .._internal.io import (
    BioimageioYamlContent,
    extract_file_name,
    interprete_file_source,
)
from ._v0_2_converter import convert_from_older_format as convert_from_older_format_v0_2


def convert_from_older_format(data: BioimageioYamlContent) -> None:
    """convert raw RDF data of an older format where possible"""
    # check if we have future format version
    fv = data.get("format_version", "0.2.0")
    if (
        not isinstance(fv, str)
        or fv.count(".") != 2
        or tuple(map(int, fv.split(".")[:2])) > (0, 3)
    ):
        return

    convert_from_older_format_v0_2(data)

    convert_attachments(data)
    convert_cover_images(data)

    _ = data.pop("download_url", None)
    _ = data.pop("rdf_source", None)

    if "name" in data and isinstance(data["name"], str):
        data["name"] = "".join(
            c if c in string.ascii_letters + string.digits + "_+- ()" else " "
            for c in data["name"]
        )[:128]

    data["format_version"] = "0.3.0"


def convert_attachments(data: BioimageioYamlContent) -> None:
    a = data.get("attachments")
    if isinstance(a, collections.abc.Mapping):
        data["attachments"] = tuple({"source": file} for file in a.get("files", []))  # type: ignore


def convert_cover_images(data: BioimageioYamlContent) -> None:
    covers = data.get("covers")
    if not isinstance(covers, list):
        return

    for i in range(len(covers)):
        c = covers[i]
        if not isinstance(c, str):
            continue

        src = interprete_file_source(c)
        fname = extract_file_name(src)

        if not (fname.endswith(".tif") or fname.endswith(".tiff")):
            continue

        try:
            image = imageio.imread(c)
            c_path = (Path(".bioimageio_converter_cache") / fname).with_suffix(".png")
            imageio.imwrite(c_path, image)
            covers[i] = str(c_path.absolute())
        except Exception as e:
            logger.warning("failed to convert tif cover image: {}", e)



import inspect
import os

from loguru import logger


def set_github_warning(title: str, message: str):
    if (
        not os.getenv("GITHUB_ACTIONS")
        or (current_frame := inspect.currentframe()) is None
        or (caller_frame := current_frame.f_back) is None
    ):
        # fallback to regular logging
        logger.opt(depth=1).warning("{}: {}", title, message)
        return

    frameinfo = inspect.getframeinfo(caller_frame)
    frameinfo.lineno

    # log according to
    # https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/workflow-commands-for-github-actions#setting-a-warning-message
    print(
        f"::warning file={frameinfo.filename},"
        + f"line={frameinfo.lineno},endLine={frameinfo.lineno + 1},"
        + f"title={title}::{message}"
    )



import collections.abc

from .._internal.io import BioimageioYamlContent
from ..generic._v0_2_converter import (
    remove_doi_prefix,
    remove_gh_prefix,
    remove_slashes_from_names,
)
from ._v0_3_converter import convert_model_from_v0_3_to_0_4_0


def convert_from_older_format(data: BioimageioYamlContent) -> None:
    fv = data.get("format_version")
    if not isinstance(fv, str):
        return

    major_minor = tuple(map(int, fv.split(".")[:2]))
    if major_minor < (0, 4):
        convert_model_from_v0_3_to_0_4_0(data)
    elif major_minor > (0, 4):
        return

    if data["format_version"] == "0.4.0":
        _convert_model_from_v0_4_0_to_0_4_1(data)

    if data["format_version"] in ("0.4.1", "0.4.2", "0.4.3", "0.4.4"):
        _convert_model_from_v0_4_4_to_0_4_5(data)

    if data["format_version"] in ("0.4.5", "0.4.6"):
        remove_slashes_from_names(data)
        data["format_version"] = "0.4.7"

    if data["format_version"] in ("0.4.7", "0.4.8"):
        data["format_version"] = "0.4.9"

    if data["format_version"] == "0.4.9":
        if isinstance(config := data.get("config"), dict) and isinstance(
            bconfig := config.get("bioimageio"), dict
        ):
            if (nickname := bconfig.get("nickname")) is not None:
                data["id"] = nickname

            if (nickname_icon := bconfig.get("nickname_icon")) is not None:
                data["id_emoji"] = nickname_icon

        data["format_version"] = "0.4.10"

    remove_doi_prefix(data)
    remove_gh_prefix(data)
    # remove 'future' from config if no other than the used future entries exist
    config = data.get("config", {})
    if isinstance(config, dict) and config.get("future") == {}:
        del config["future"]

    # remove 'config' if now empty
    if data.get("config") == {}:
        del data["config"]


def _convert_model_from_v0_4_0_to_0_4_1(data: BioimageioYamlContent):
    # move dependencies from root to pytorch_state_dict weights entry
    deps = data.pop("dependencies", None)
    weights = data.get("weights", {})
    if deps and weights and isinstance(weights, dict):
        entry = weights.get("pytorch_state_dict")
        if entry and isinstance(entry, dict):
            entry["dependencies"] = deps

    data["format_version"] = "0.4.1"


def _convert_model_from_v0_4_4_to_0_4_5(data: BioimageioYamlContent) -> None:
    parent = data.pop("parent", None)
    if isinstance(parent, collections.abc.Mapping) and "uri" in parent:
        data["parent"] = parent["uri"]

    data["format_version"] = "0.4.5"



import collections.abc
from typing import Any, Dict, Mapping, Union

from .._internal.io import BioimageioYamlContent
from .._internal.type_guards import is_mapping


def convert_from_older_format(data: BioimageioYamlContent) -> None:
    """convert raw RDF data of an older format where possible"""
    # check if we have future format version
    if "format_version" not in data:
        return

    fv = data["format_version"]
    if isinstance(fv, str) and tuple(map(int, fv.split(".")[:2])) > (0, 2):
        return

    # we unofficially accept strings as author entries
    authors = data.get("authors")
    if isinstance(authors, list):
        data["authors"] = [{"name": a} if isinstance(a, str) else a for a in authors]

    if data.get("format_version") in ("0.2.0", "0.2.1"):
        data["format_version"] = "0.2.2"

    if data.get("format_version") == "0.2.2":
        remove_slashes_from_names(data)
        data["format_version"] = "0.2.3"

    if data.get("format_version") == "0.2.3":
        if isinstance(config := data.get("config"), dict) and isinstance(
            bconfig := config.get("bioimageio"), dict
        ):
            if (nickname := bconfig.get("nickname")) is not None:
                data["id"] = nickname

            if (nickname_icon := bconfig.get("nickname_icon")) is not None:
                data["id_emoji"] = nickname_icon

        data["format_version"] = "0.2.4"

    remove_doi_prefix(data)
    remove_gh_prefix(data)


def remove_slashes_from_names(data: Dict[Any, Any]) -> None:
    NAME = "name"
    if NAME in data and isinstance(data[NAME], str):
        data[NAME] = data[NAME].replace("/", "").replace("\\", "")

    # update authors and maintainers
    def rm_slashes_in_person_name(
        person: Union[Any, Mapping[Union[Any, str], Any]],
    ) -> Any:
        if not is_mapping(person):
            return person

        new_person = dict(person)
        if isinstance(n := person.get(NAME), str):
            new_person[NAME] = n.replace("/", "").replace("\\", "")

        return new_person

    for group in ("authors", "maintainers"):
        persons = data.get(group)
        if isinstance(persons, collections.abc.Sequence):
            data[group] = [rm_slashes_in_person_name(p) for p in persons]  # type: ignore


DOI_PREFIXES = ("https://doi.org/", "http://dx.doi.org/")


def remove_doi_prefix(data: BioimageioYamlContent) -> None:
    """we unofficially accept DOIs starting with "https://doi.org/" here we remove this prefix"""
    cite = data.get("cite")
    if isinstance(cite, collections.abc.Sequence):
        new_cite = list(cite)
        for i in range(len(new_cite)):
            cite_entry = new_cite[i]
            if not isinstance(cite_entry, collections.abc.Mapping):
                continue

            doi = cite_entry.get("doi")
            if not isinstance(doi, str):
                continue

            for doi_prefix in DOI_PREFIXES:
                if doi.startswith(doi_prefix):
                    doi = doi[len(doi_prefix) :]
                    break
            else:
                continue

            new_cite_entry = dict(cite_entry)
            new_cite_entry["doi"] = doi
            new_cite[i] = new_cite_entry

        data["cite"] = new_cite


def remove_gh_prefix(data: BioimageioYamlContent) -> None:
    def rm_gh(field_name: str):
        authors = data.get(field_name)
        if not isinstance(authors, list):
            return

        for a in authors:
            if (
                isinstance(a, dict)
                and "github_user" in a
                and isinstance(a["github_user"], str)
                and a["github_user"].startswith("https://github.com/")
            ):
                a["github_user"] = a["github_user"][len("https://github.com/") :]

    rm_gh("authors")
    rm_gh("maintainers")



# type: ignore
from typing import Any, Dict


def convert_model_from_v0_3_to_0_4_0(data: Dict[str, Any]) -> None:
    """auto converts model 'data' to newest format"""

    if "format_version" not in data:
        return

    if data["format_version"] == "0.3.0":
        # no breaking change, bump to 0.3.1
        data["format_version"] = "0.3.1"

    if data["format_version"] == "0.3.1":
        data = _convert_model_v0_3_1_to_v0_3_2(data)

    if data["format_version"] == "0.3.2":
        data = _convert_model_v0_3_2_to_v0_3_3(data)

    if data["format_version"] in ("0.3.3", "0.3.4", "0.3.5"):
        data["format_version"] = "0.3.6"

    if data["format_version"] != "0.3.6":
        return

    # remove 'future' from config if no other than the used future entries exist
    config = data.get("config", {})
    if config.get("future") == {}:
        del config["future"]

    # remove 'config' if now empty
    if data.get("config") == {}:
        del data["config"]

    data.pop("language", None)
    data.pop("framework", None)

    architecture = data.pop("source", None)
    architecture_sha256 = data.pop("sha256", None)
    kwargs = data.pop("kwargs", None)
    pytorch_state_dict_weights_entry = data.get("weights", {}).get("pytorch_state_dict")
    if pytorch_state_dict_weights_entry is not None:
        if architecture is not None:
            pytorch_state_dict_weights_entry["architecture"] = architecture

        if architecture_sha256 is not None:
            pytorch_state_dict_weights_entry["architecture_sha256"] = (
                architecture_sha256
            )

        if kwargs is not None:
            pytorch_state_dict_weights_entry["kwargs"] = kwargs

    torchscript_weights_entry = data.get("weights", {}).pop("pytorch_script", None)
    if torchscript_weights_entry is not None:
        data.setdefault("weights", {})["torchscript"] = torchscript_weights_entry

    data["format_version"] = "0.4.0"


def _convert_model_v0_3_1_to_v0_3_2(data: Dict[str, Any]) -> Dict[str, Any]:
    data["type"] = "model"
    data["format_version"] = "0.3.2"
    future = data.get("config", {}).get("future", {}).pop("0.3.2", {})

    authors = data.get("authors")
    if isinstance(authors, list):
        data["authors"] = [{"name": name} for name in authors]
        authors_update = future.get("authors")
        if authors_update is not None:
            for a, u in zip(data["authors"], authors_update):
                a.update(u)

    # packaged_by
    packaged_by = data.get("packaged_by")
    if packaged_by is not None:
        data["packaged_by"] = [{"name": name} for name in data["packaged_by"]]
        packaged_by_update = future.get("packaged_by")
        if packaged_by_update is not None:
            for a, u in zip(data["packaged_by"], packaged_by_update):
                a.update(u)

    # authors of weights
    weights = data.get("weights")
    if isinstance(weights, dict):
        for weights_format, weights_entry in weights.items():
            if "authors" not in weights_entry:
                continue

            weights_entry["authors"] = [
                {"name": name} for name in weights_entry["authors"]
            ]
            authors_update = (
                future.get("weights", {}).get(weights_format, {}).get("authors")
            )
            if authors_update is not None:
                for a, u in zip(weights_entry["authors"], authors_update):
                    a.update(u)

    # model version
    if "version" in future:
        data["version"] = future.pop("version")

    return data


def _convert_model_v0_3_2_to_v0_3_3(data: Dict[str, Any]) -> Dict[str, Any]:
    data["format_version"] = "0.3.3"
    if "outputs" in data:
        for out in data["outputs"]:
            if "shape" in out:
                shape = out["shape"]
                if isinstance(shape, dict) and "reference_input" in shape:
                    shape["reference_tensor"] = shape.pop("reference_input")

    return data



from pathlib import Path
from typing import Any, Iterable, Mapping

import pytest

from tests.conftest import EXAMPLE_DESCRIPTIONS
from tests.utils import ParameterSet, check_bioimageio_yaml


def get_param(descr_path: Path) -> ParameterSet:
    key = descr_path.relative_to(EXAMPLE_DESCRIPTIONS).as_posix()
    return pytest.param(descr_path, key, id=key)


def yield_valid_descr_paths() -> Iterable[ParameterSet]:
    assert EXAMPLE_DESCRIPTIONS.exists()
    for p in EXAMPLE_DESCRIPTIONS.glob("**/*bioimageio.yaml"):
        if p.name.startswith("invalid"):
            continue

        yield get_param(p)


def yield_invalid_descr_paths() -> Iterable[ParameterSet]:
    assert EXAMPLE_DESCRIPTIONS.exists()
    for p in EXAMPLE_DESCRIPTIONS.glob("**/invalid*bioimageio.yaml"):
        yield get_param(p)


EXCLUDE_FIELDS_FROM_ROUNDTRIP = {
    "models/stardist_example_model/v0_4.bioimageio.yaml": {"dependencies"},
    "models/stardist_example_model/wrong_shape_v0_4.bioimageio.yaml": {"dependencies"},
    "models/stardist_example_model/wrong_shape2_v0_4.bioimageio.yaml": {"dependencies"},
    # "models/unet2d_diff_output_shape/bioimageio.yaml": {
    #     "dependencies",
    #     "weights",
    # },
    "models/unet2d_multi_tensor/bioimageio.yaml": {"dependencies", "weights"},
    "models/unet2d_nuclei_broad/v0_4_0.bioimageio.yaml": {
        "dependencies",
        "weights",
        "version",
    },
    "models/unet2d_nuclei_broad/v0_4_9.bioimageio.yaml": {"version"},
    "models/upsample_test_model/v0_4.bioimageio.yaml": {"dependencies", "weights"},
}


@pytest.mark.parametrize("descr_path,key", list(yield_valid_descr_paths()))
def test_example_descr_paths(
    descr_path: Path,
    key: str,
    bioimageio_json_schema: Mapping[Any, Any],
):
    check_bioimageio_yaml(
        descr_path,
        root=descr_path.parent,
        as_latest=False,
        exclude_fields_from_roundtrip=EXCLUDE_FIELDS_FROM_ROUNDTRIP.get(key, set()),
        bioimageio_json_schema=bioimageio_json_schema,
        perform_io_checks=False,
    )


@pytest.mark.parametrize("descr_path,key", list(yield_invalid_descr_paths()))
def test_invalid_example_descr_paths(
    descr_path: Path, key: str, bioimageio_json_schema: Mapping[Any, Any]
):
    check_bioimageio_yaml(
        descr_path,
        root=descr_path.parent,
        as_latest=False,
        is_invalid=True,
        bioimageio_json_schema=bioimageio_json_schema,
        perform_io_checks=False,
    )



from bioimageio.spec import ValidationContext, update_format, validate_format
from bioimageio.spec.common import BioimageioYamlContent, RootHttpUrl
from bioimageio.spec.model import ModelDescr

EXAMPLE_COM = RootHttpUrl("https://example.com/")


def test_forward_compatibility(unet2d_data: BioimageioYamlContent):
    data = dict(unet2d_data)
    v_future = "0.9999.0"
    data["format_version"] = v_future  # assume it is valid in a future format version

    summary = validate_format(
        data,
        context=ValidationContext(root=EXAMPLE_COM, perform_io_checks=False),
    )
    assert summary.status == "valid-format", summary.errors

    # expect warning about treating future format version as latest
    ws = summary.warnings
    assert len(ws) >= 1, ws
    assert ws[0].msg.startswith("future format_version '0.9999.0' treated as ")


def test_no_forward_compatibility(unet2d_data: BioimageioYamlContent):
    data = dict(unet2d_data)
    data["authors"] = 42  # make sure rdf is invalid
    data["format_version"] = "0.9999.0"  # assume it is valid in a future format version

    summary = validate_format(
        data,
        context=ValidationContext(root=EXAMPLE_COM, perform_io_checks=False),
    )
    assert summary.status == "failed", summary

    assert len(summary.errors) == 1, summary.errors
    assert summary.errors[0].loc == ("authors",), summary.errors[0].loc

    # expect warning about treating future format version as latest
    ws = summary.warnings
    assert len(ws) >= 1, ws
    assert ws[0].msg.startswith("future format_version '0.9999.0' treated as ")


def test_update_format(unet2d_path_old: str):
    updated = update_format(unet2d_path_old, perform_io_checks=False)
    assert updated.type == "model"
    assert updated.format_version == ModelDescr.implemented_format_version



from pathlib import Path

import pytest

from bioimageio.spec import load_model_description
from bioimageio.spec.model.v0_5 import (
    DatasetId,
    Evaluation,
    HttpUrl,
    LicenseId,
    LinkedModel,
    ModelId,
    OnnxWeightsDescr,
    Version,
    WeightsDescr,
)


@pytest.mark.parametrize("fill_meta", (True, False))
def test_hf(unet2d_path: Path, fill_meta: bool, tmp_path: Path):
    from bioimageio.spec._hf import push_to_hub

    model = load_model_description(unet2d_path, format_version="latest")
    model.id = ModelId("unet2d_nuclei_broad")

    if fill_meta:
        model.license = LicenseId("AAL")
        model.parent = LinkedModel(
            id=ModelId("unet2d_nuclei_broad_v0.1"),
            version=Version("0.1"),
        )
        model.config.bioimageio.architecture_type = "Unet"
        model.config.bioimageio.architecture_description = "A 2D Unet"
        model.config.bioimageio.modality = "fluorescence microscopy"
        model.config.bioimageio.target_structure = ["nuclei"]
        model.config.bioimageio.task = "segmentation"
        model.config.bioimageio.environmental_impact.hardware_type = "GTX 2080 Ti"
        model.config.bioimageio.environmental_impact.hours_used = 10
        model.config.bioimageio.environmental_impact.compute_region = "Germany"
        model.config.bioimageio.environmental_impact.co2_emitted = 0.54
        model.config.bioimageio.environmental_impact.cloud_provider = "EMBL Heidelberg"
        model.config.bioimageio.funded_by = "EMBL"
        model.config.bioimageio.new_version = ModelId(
            "huggingface/bioimageio/not_a_real_model"
        )
        model.config.bioimageio.out_of_scope_use = (
            "for testing only. do not use this model for real applications."
        )
        model.config.bioimageio.model_parameter_count = 123456
        model.config.bioimageio.training.training_epochs = 100
        model.config.bioimageio.training.training_batch_size = 10
        model.config.bioimageio.training.initial_learning_rate = 1e-4
        model.config.bioimageio.training.learning_rate_schedule = "ausgefuxt"
        model.config.bioimageio.training.loss_function = "dice_loss"
        model.config.bioimageio.training.loss_function_kwargs = {"smooth": 1e-5}
        model.config.bioimageio.training.optimizer = "Adam"
        model.config.bioimageio.training.optimizer_kwargs = {"weight_decay": 1e-5}
        model.config.bioimageio.training.regularization = "drop-out"
        model.config.bioimageio.training.training_duration = 1
        model.config.bioimageio.training.training_preprocessing = (
            "some preprocessing description"
        )
        model.config.bioimageio.inference_time = "super fast"
        model.config.bioimageio.memory_requirements_inference = "not much"
        model.config.bioimageio.memory_requirements_training = "also not much"
        model.config.bioimageio.evaluations.append(
            Evaluation(
                dataset_id=DatasetId("my_test_set"),
                dataset_source=HttpUrl("https://example.com/dataset_source"),
                dataset_role="test",
                sample_count=3,
                evaluation_factors=["nuclei", "membrane"],
                evaluation_factors_long=["cell nuclei", "cell membrane"],
                metrics=["DICE", "IoU"],
                metrics_long=["dice score", "IoU score"],
                results=[[1, 2], [3, 4]],
                results_summary="dummy summary",
            ),
        )
        model.config.bioimageio.evaluations.append(
            Evaluation(
                model_id=model.id,
                dataset_id=DatasetId("my_test_set"),
                dataset_source=HttpUrl("https://example.com/dataset_source"),
                dataset_role="independent",
                sample_count=3,
                evaluation_factors=["nuclei", "membrane"],
                evaluation_factors_long=["cell nuclei", "cell membrane"],
                metrics=["DICE", "IoU"],
                metrics_long=["dice score", "IoU score"],
                results=[[1, 2], [3, 4]],
                results_summary="dummy summary",
            ),
        )
    else:
        model.authors = []
        model.cite = []
        model.license = None
        assert model.weights.onnx is not None
        model.weights = WeightsDescr(
            onnx=OnnxWeightsDescr(
                source=model.weights.onnx.source,
                opset_version=11,
                comment="fake example weights",
            )
        )  # create invalid empty weights descr

    push_to_hub(model, "thefynnbe", prep_only_no_upload=True, prep_dir=tmp_path)

    if not fill_meta:
        with pytest.raises(ValueError):  # non-empty output dir
            push_to_hub(model, "thefynnbe", prep_only_no_upload=True, prep_dir=tmp_path)

        model.id = None
        with pytest.raises(ValueError):
            push_to_hub(model, "thefynnbe", prep_only_no_upload=True)



# autogen: start
"""notebook resource description format

Implementations of all released minor versions are available in submodules.
"""

from typing import Union

from pydantic import Discriminator, Field
from typing_extensions import Annotated

from . import v0_2, v0_3

NotebookDescr = v0_3.NotebookDescr
NotebookDescr_v0_2 = v0_2.NotebookDescr
NotebookDescr_v0_3 = v0_3.NotebookDescr

AnyNotebookDescr = Annotated[
    Union[
        Annotated[NotebookDescr_v0_2, Field(title="notebook 0.2")],
        Annotated[NotebookDescr_v0_3, Field(title="notebook 0.3")],
    ],
    Discriminator("format_version"),
    Field(title="notebook"),
]
"""Union of any released notebook desription"""
# autogen: stop



import collections.abc
import io
from typing import Union
from zipfile import ZipFile

import httpx
from loguru import logger

from bioimageio.spec._io import load_description

from ._description import (
    InvalidDescr,
    ResourceDescr,
    build_description,
)
from ._internal._settings import settings
from ._internal.common_nodes import ResourceDescrBase
from ._internal.io import BioimageioYamlContent, get_reader
from ._internal.io_basics import BIOIMAGEIO_YAML
from ._internal.io_utils import write_yaml
from ._internal.validation_context import get_validation_context
from ._package import get_resource_package_content
from .common import HttpUrl, PermissiveFileSource


# TODO: remove alpha stage warning
def upload(
    source: Union[PermissiveFileSource, ZipFile, ResourceDescr, BioimageioYamlContent],
    /,
) -> HttpUrl:
    """Upload a new resource description (version) to the hypha server to be shared at bioimage.io.
    To edit an existing resource **version**, please login to https://bioimage.io and use the web interface.

    WARNING: This upload function is in alpha stage and might change in the future.

    Args:
        source: The resource description to upload.

    Returns:
        A URL to the uploaded resource description.
        Note: It might take some time until the resource is processed and available for download from the returned URL.
    """

    if settings.hypha_upload_token is None:
        raise ValueError(
            """
Upload token is not set. Please set BIOIMAGEIO_HYPHA_UPLOAD_TOKEN in your environment variables.
By setting this token you agree to our terms of service at https://bioimage.io/#/toc.

How to obtain a token:
    1. Login to https://bioimage.io
    2. Generate a new token at https://bioimage.io/#/api?tab=hypha-rpc
"""
        )

    if isinstance(source, ResourceDescrBase):
        # If source is already a ResourceDescr, we can use it directly
        descr = source
    elif isinstance(source, dict):
        descr = build_description(source)
    else:
        descr = load_description(source)

    if isinstance(descr, InvalidDescr):
        raise ValueError("Uploading invalid resource descriptions is not allowed.")

    if descr.type != "model":
        raise NotImplementedError(
            f"For now, only model resources can be uploaded (got type={descr.type})."
        )

    if descr.id is not None:
        raise ValueError(
            "You cannot upload a resource with an id. Please remove the id from the description and make sure to upload a new non-existing resource. To edit an existing resource, please use the web interface at https://bioimage.io."
        )

    content = get_resource_package_content(descr)

    metadata = content[BIOIMAGEIO_YAML]
    assert isinstance(metadata, dict)
    manifest = dict(metadata)

    # only admins can upload a resource with a version
    artifact_version = "stage"  # if descr.version is None else str(descr.version)

    # Create new model
    r = httpx.post(
        settings.hypha_upload,
        json={
            "parent_id": "bioimage-io/bioimage.io",
            "alias": (
                descr.id or "{animal_adjective}-{animal}"
            ),  # TODO: adapt for non-model uploads,
            "type": descr.type,
            "manifest": manifest,
            "version": artifact_version,
        },
        headers=(
            headers := {
                "Authorization": f"Bearer {settings.hypha_upload_token}",
                "Content-Type": "application/json",
            }
        ),
    )

    response = r.json()
    artifact_id = response.get("id")
    if artifact_id is None:
        try:
            logger.error("Response detail: {}", "".join(response["detail"]))
        except Exception:
            logger.error("Response: {}", response)

        raise RuntimeError(f"Upload did not return resource id: {response}")
    else:
        logger.info("Uploaded resource description {}", artifact_id)

    for file_name, file_source in content.items():
        # Get upload URL for a file
        response = httpx.post(
            settings.hypha_upload.replace("/create", "/put_file"),
            json={
                "artifact_id": artifact_id,
                "file_path": file_name,
            },
            headers=headers,
            follow_redirects=True,
        )
        upload_url = response.raise_for_status().json()

        # Upload file to the provided URL
        if isinstance(file_source, collections.abc.Mapping):
            buf = io.BytesIO()
            write_yaml(file_source, buf)
            files = {file_name: buf}
        else:
            files = {file_name: get_reader(file_source)}

        response = httpx.put(
            upload_url,
            files=files,  # pyright: ignore[reportArgumentType]
            # TODO: follow up on https://github.com/encode/httpx/discussions/3611
            headers={"Content-Type": ""},  # Important for S3 uploads
            follow_redirects=True,
        )
        logger.info("Uploaded '{}' successfully", file_name)

    # Update model status
    manifest["status"] = "request-review"
    response = httpx.post(
        settings.hypha_upload.replace("/create", "/edit"),
        json={
            "artifact_id": artifact_id,
            "version": artifact_version,
            "manifest": manifest,
        },
        headers=headers,
        follow_redirects=True,
    )
    logger.info(
        "Updated status of {}/{} to 'request-review'", artifact_id, artifact_version
    )
    logger.warning(
        "Upload successfull. Please note that the uploaded resource might not be available for download immediately."
    )
    with get_validation_context().replace(perform_io_checks=False):
        return HttpUrl(
            f"https://hypha.aicell.io/bioimage-io/artifacts/{artifact_id}/files/rdf.yaml?version={artifact_version}"
        )



import collections.abc
import shutil
from io import BytesIO
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from typing import IO, Dict, Literal, Optional, Sequence, Union
from zipfile import ZIP_DEFLATED

from loguru import logger
from pydantic import DirectoryPath, FilePath, NewPath

from ._description import InvalidDescr, ResourceDescr, build_description
from ._internal.common_nodes import ResourceDescrBase
from ._internal.io import (
    BioimageioYamlContent,
    BioimageioYamlSource,
    FileDescr,
    RelativeFilePath,
    ensure_is_valid_bioimageio_yaml_name,
)
from ._internal.io_basics import (
    BIOIMAGEIO_YAML,
    AbsoluteFilePath,
    BytesReader,
    FileName,
    ZipPath,
)
from ._internal.io_utils import open_bioimageio_yaml, write_yaml, write_zip
from ._internal.packaging_context import PackagingContext
from ._internal.url import HttpUrl
from ._internal.utils import get_os_friendly_file_name
from ._internal.validation_context import get_validation_context
from ._internal.warning_levels import ERROR
from ._io import load_description
from .model.v0_4 import WeightsFormat


# TODO: deprecate in favor of get_package_content
def get_resource_package_content(
    rd: ResourceDescr,
    /,
    *,
    bioimageio_yaml_file_name: FileName = BIOIMAGEIO_YAML,
    weights_priority_order: Optional[Sequence[WeightsFormat]] = None,  # model only
) -> Dict[FileName, Union[HttpUrl, AbsoluteFilePath, BioimageioYamlContent, ZipPath]]:
    """Get the content of a bioimage.io resource package."""
    ret: Dict[
        FileName, Union[HttpUrl, AbsoluteFilePath, BioimageioYamlContent, ZipPath]
    ] = {}
    for k, v in get_package_content(
        rd,
        bioimageio_yaml_file_name=bioimageio_yaml_file_name,
        weights_priority_order=weights_priority_order,
    ).items():
        if isinstance(v, FileDescr):
            if isinstance(v.source, (Path, RelativeFilePath)):
                ret[k] = v.source.absolute()
            else:
                ret[k] = v.source

        else:
            ret[k] = v

    return ret


def get_package_content(
    rd: ResourceDescr,
    /,
    *,
    bioimageio_yaml_file_name: FileName = BIOIMAGEIO_YAML,
    weights_priority_order: Optional[Sequence[WeightsFormat]] = None,  # model only
) -> Dict[FileName, Union[FileDescr, BioimageioYamlContent]]:
    """
    Args:
        rd: resource description
        bioimageio_yaml_file_name: RDF file name
        weights_priority_order: (for model resources only)
            If given, only the first weights format present in the model is included.
            If none of the prioritized weights formats is found a ValueError is raised.
    """
    os_friendly_name = get_os_friendly_file_name(rd.name)
    bioimageio_yaml_file_name = bioimageio_yaml_file_name.format(
        name=os_friendly_name, type=rd.type
    )

    bioimageio_yaml_file_name = ensure_is_valid_bioimageio_yaml_name(
        bioimageio_yaml_file_name
    )
    content: Dict[FileName, FileDescr] = {}
    with PackagingContext(
        bioimageio_yaml_file_name=bioimageio_yaml_file_name,
        file_sources=content,
        weights_priority_order=weights_priority_order,
    ):
        rdf_content: BioimageioYamlContent = rd.model_dump(
            mode="json", exclude_unset=True
        )

    _ = rdf_content.pop("rdf_source", None)

    return {**content, bioimageio_yaml_file_name: rdf_content}


def _prepare_resource_package(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    weights_priority_order: Optional[Sequence[WeightsFormat]] = None,
) -> Dict[FileName, Union[BioimageioYamlContent, BytesReader]]:
    """Prepare to package a resource description; downloads all required files.

    Args:
        source: A bioimage.io resource description (as file, raw YAML content or description class)
        context: validation context
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.
    """
    context = get_validation_context()
    bioimageio_yaml_file_name = context.file_name
    if isinstance(source, ResourceDescrBase):
        descr = source
    elif isinstance(source, collections.abc.Mapping):
        descr = build_description(source)
    else:
        opened = open_bioimageio_yaml(source)
        bioimageio_yaml_file_name = opened.original_file_name
        context = context.replace(
            root=opened.original_root, file_name=opened.original_file_name
        )
        with context:
            descr = build_description(opened.content)

    if isinstance(descr, InvalidDescr):
        raise ValueError(f"{source} is invalid: {descr.validation_summary}")

    with context:
        package_content = get_package_content(
            descr,
            bioimageio_yaml_file_name=bioimageio_yaml_file_name or BIOIMAGEIO_YAML,
            weights_priority_order=weights_priority_order,
        )

    return {
        k: v if isinstance(v, collections.abc.Mapping) else v.get_reader()
        for k, v in package_content.items()
    }


def save_bioimageio_package_as_folder(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    output_path: Union[NewPath, DirectoryPath, None] = None,
    weights_priority_order: Optional[  # model only
        Sequence[
            Literal[
                "keras_hdf5",
                "onnx",
                "pytorch_state_dict",
                "tensorflow_js",
                "tensorflow_saved_model_bundle",
                "torchscript",
            ]
        ]
    ] = None,
) -> DirectoryPath:
    """Write the content of a bioimage.io resource package to a folder.

    Args:
        source: bioimageio resource description
        output_path: file path to write package to
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Returns:
        directory path to bioimageio package folder
    """
    package_content = _prepare_resource_package(
        source,
        weights_priority_order=weights_priority_order,
    )
    if output_path is None:
        output_path = Path(mkdtemp())
    else:
        output_path = Path(output_path)

    output_path.mkdir(exist_ok=True, parents=True)
    for name, src in package_content.items():
        if not name:
            raise ValueError("got empty file name in package content")

        if isinstance(src, collections.abc.Mapping):
            write_yaml(src, output_path / name)
        elif (
            isinstance(src.original_root, Path)
            and src.original_root / src.original_file_name
            == (output_path / name).resolve()
        ):
            logger.debug(
                f"Not copying {src.original_root / src.original_file_name} to itself."
            )
        else:
            if isinstance(src.original_root, Path):
                logger.debug(
                    f"Copying from path {src.original_root / src.original_file_name} to {output_path / name}."
                )
            else:
                logger.debug(
                    f"Copying {src.original_root}/{src.original_file_name} to {output_path / name}."
                )
            with (output_path / name).open("wb") as dest:
                _ = shutil.copyfileobj(src, dest)

    return output_path


def save_bioimageio_package(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    compression: int = ZIP_DEFLATED,
    compression_level: int = 1,
    output_path: Union[NewPath, FilePath, None] = None,
    weights_priority_order: Optional[  # model only
        Sequence[
            Literal[
                "keras_hdf5",
                "onnx",
                "pytorch_state_dict",
                "tensorflow_js",
                "tensorflow_saved_model_bundle",
                "torchscript",
            ]
        ]
    ] = None,
    allow_invalid: bool = False,
) -> FilePath:
    """Package a bioimageio resource as a zip file.

    Args:
        source: bioimageio resource description
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile
        output_path: file path to write package to
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Returns:
        path to zipped bioimageio package
    """
    package_content = _prepare_resource_package(
        source,
        weights_priority_order=weights_priority_order,
    )
    if output_path is None:
        output_path = Path(
            NamedTemporaryFile(suffix=".bioimageio.zip", delete=False).name
        )
    else:
        output_path = Path(output_path)

    write_zip(
        output_path,
        package_content,
        compression=compression,
        compression_level=compression_level,
    )
    with get_validation_context().replace(warning_level=ERROR):
        if isinstance((exported := load_description(output_path)), InvalidDescr):
            exported.validation_summary.display()
            msg = f"Exported package at '{output_path}' is invalid."
            if allow_invalid:
                logger.error(msg)
            else:
                raise ValueError(msg)

    return output_path


def save_bioimageio_package_to_stream(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    compression: int = ZIP_DEFLATED,
    compression_level: int = 1,
    output_stream: Union[IO[bytes], None] = None,
    weights_priority_order: Optional[  # model only
        Sequence[
            Literal[
                "keras_hdf5",
                "onnx",
                "pytorch_state_dict",
                "tensorflow_js",
                "tensorflow_saved_model_bundle",
                "torchscript",
            ]
        ]
    ] = None,
) -> IO[bytes]:
    """Package a bioimageio resource into a stream.

    Args:
        source: bioimageio resource description
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile
        output_stream: stream to write package to
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Note: this function bypasses safety checks and does not load/validate the model after writing.

    Returns:
        stream of zipped bioimageio package
    """
    if output_stream is None:
        output_stream = BytesIO()

    package_content = _prepare_resource_package(
        source,
        weights_priority_order=weights_priority_order,
    )

    write_zip(
        output_stream,
        package_content,
        compression=compression,
        compression_level=compression_level,
    )

    return output_stream



from __future__ import annotations

from contextvars import ContextVar, Token
from copy import copy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Dict, List, Literal, Optional, Union, cast
from urllib.parse import urlsplit, urlunsplit
from zipfile import ZipFile

from genericache import DiskCache, MemoryCache, NoopCache
from pydantic import ConfigDict, DirectoryPath
from typing_extensions import Self

from ._settings import settings
from .io_basics import FileName, Sha256
from .progress import Progressbar
from .root_url import RootHttpUrl
from .utils import SLOTS
from .warning_levels import WarningLevel


@dataclass(frozen=True, **SLOTS)
class ValidationContextBase:
    file_name: Optional[FileName] = None
    """File name of the bioimageio YAML file."""

    original_source_name: Optional[str] = None
    """Original source of the bioimageio resource description, e.g. a URL or file path."""

    perform_io_checks: bool = settings.perform_io_checks
    """Wether or not to perform validation that requires file io,
    e.g. downloading a remote files.

    Existence of local absolute file paths is still being checked."""

    known_files: Dict[str, Optional[Sha256]] = field(
        default_factory=cast(  # TODO: (py>3.8) use dict[str, Optional[Sha256]]
            Callable[[], Dict[str, Optional[Sha256]]], dict
        )
    )
    """Allows to bypass download and hashing of referenced files."""

    update_hashes: bool = False
    """Overwrite specified file hashes with values computed from the referenced file (instead of comparing them).
    (Has no effect if `perform_io_checks=False`.)"""


@dataclass(frozen=True)
class ValidationContextSummary(ValidationContextBase):
    """Summary of the validation context without internally used context fields."""

    __pydantic_config__ = ConfigDict(extra="forbid")
    """Pydantic config to include **ValdationContextSummary** in **ValidationDetail**."""

    root: Union[RootHttpUrl, Path, Literal["in-memory"]] = Path()


@dataclass(frozen=True)
class ValidationContext(ValidationContextBase):
    """A validation context used to control validation of bioimageio resources.

    For example a relative file path in a bioimageio description requires the **root**
    context to evaluate if the file is available and, if **perform_io_checks** is true,
    if it matches its expected SHA256 hash value.
    """

    _context_tokens: "List[Token[Optional[ValidationContext]]]" = field(
        init=False,
        default_factory=cast(
            "Callable[[], List[Token[Optional[ValidationContext]]]]", list
        ),
    )

    cache: Union[
        DiskCache[RootHttpUrl], MemoryCache[RootHttpUrl], NoopCache[RootHttpUrl]
    ] = field(default=settings.disk_cache)
    disable_cache: bool = False
    """Disable caching downloads to `settings.cache_path`
    and (re)download them to memory instead."""

    root: Union[RootHttpUrl, DirectoryPath, ZipFile] = Path()
    """Url/directory/archive serving as base to resolve any relative file paths."""

    warning_level: WarningLevel = 50
    """Treat warnings of severity `s` as validation errors if `s >= warning_level`."""

    log_warnings: bool = settings.log_warnings
    """If `True` warnings are logged to the terminal

    Note:
        This setting does not affect warning entries
        of a generated `bioimageio.spec.ValidationSummary`.
    """

    progressbar: Union[None, bool, Callable[[], Progressbar]] = None
    """Control any progressbar.
    (Currently this is only used for file downloads.)

    Can be:
    - `None`: use a default tqdm progressbar (if not settings.CI)
    - `True`: use a default tqdm progressbar
    - `False`: disable the progressbar
    - `callable`: A callable that returns a tqdm-like progressbar.
    """

    raise_errors: bool = False
    """Directly raise any validation errors
    instead of aggregating errors and returning a `bioimageio.spec.InvalidDescr`. (for debugging)"""

    @property
    def summary(self):
        if isinstance(self.root, ZipFile):
            if self.root.filename is None:
                root = "in-memory"
            else:
                root = Path(self.root.filename)
        else:
            root = self.root

        return ValidationContextSummary(
            root=root,
            file_name=self.file_name,
            perform_io_checks=self.perform_io_checks,
            known_files=copy(self.known_files),
            update_hashes=self.update_hashes,
        )

    def __enter__(self):
        self._context_tokens.append(_validation_context_var.set(self))
        return self

    def __exit__(self, type, value, traceback):  # type: ignore
        _validation_context_var.reset(self._context_tokens.pop(-1))

    def replace(  # TODO: probably use __replace__ when py>=3.13
        self,
        root: Optional[Union[RootHttpUrl, DirectoryPath, ZipFile]] = None,
        warning_level: Optional[WarningLevel] = None,
        log_warnings: Optional[bool] = None,
        file_name: Optional[str] = None,
        perform_io_checks: Optional[bool] = None,
        known_files: Optional[Dict[str, Optional[Sha256]]] = None,
        raise_errors: Optional[bool] = None,
        update_hashes: Optional[bool] = None,
        original_source_name: Optional[str] = None,
    ) -> Self:
        if known_files is None and root is not None and self.root != root:
            # reset known files if root changes, but no new known_files are given
            known_files = {}

        return self.__class__(
            root=self.root if root is None else root,
            warning_level=(
                self.warning_level if warning_level is None else warning_level
            ),
            log_warnings=self.log_warnings if log_warnings is None else log_warnings,
            file_name=self.file_name if file_name is None else file_name,
            perform_io_checks=(
                self.perform_io_checks
                if perform_io_checks is None
                else perform_io_checks
            ),
            known_files=self.known_files if known_files is None else known_files,
            raise_errors=self.raise_errors if raise_errors is None else raise_errors,
            update_hashes=(
                self.update_hashes if update_hashes is None else update_hashes
            ),
            original_source_name=(
                self.original_source_name
                if original_source_name is None
                else original_source_name
            ),
        )

    @property
    def source_name(self) -> str:
        if self.original_source_name is not None:
            return self.original_source_name
        elif self.file_name is None:
            return "in-memory"
        else:
            try:
                if isinstance(self.root, Path):
                    source = (self.root / self.file_name).absolute()
                else:
                    parsed = urlsplit(str(self.root))
                    path = list(parsed.path.strip("/").split("/")) + [self.file_name]
                    source = urlunsplit(
                        (
                            parsed.scheme,
                            parsed.netloc,
                            "/".join(path),
                            parsed.query,
                            parsed.fragment,
                        )
                    )
            except ValueError:
                return self.file_name
            else:
                return str(source)


_validation_context_var: ContextVar[Optional[ValidationContext]] = ContextVar(
    "validation_context_var", default=None
)


def get_validation_context(
    default: Optional[ValidationContext] = None,
) -> ValidationContext:
    """Get the currently active validation context (or a default)"""
    return _validation_context_var.get() or default or ValidationContext()



from dataclasses import dataclass
from typing import Any, Type

import annotated_types
from pydantic import GetCoreSchemaHandler, functional_validators
from pydantic_core import CoreSchema
from pydantic_core.core_schema import no_info_after_validator_function

from .utils import SLOTS


# TODO: make sure we use this one everywhere and not the vanilla pydantic one
@dataclass(frozen=True, **SLOTS)
class AfterValidator(functional_validators.AfterValidator):
    def __str__(self):
        return f"AfterValidator({self.func.__name__})"


# TODO: make sure we use this one everywhere and not the vanilla pydantic one
@dataclass(frozen=True, **SLOTS)
class BeforeValidator(functional_validators.BeforeValidator):
    def __str__(self):
        return f"BeforeValidator({self.func.__name__})"


# TODO: make sure we use this one everywhere and not the vanilla pydantic one
@dataclass(frozen=True, **SLOTS)
class Predicate(annotated_types.Predicate):
    def __str__(self):
        return f"Predicate({self.func.__name__})"


@dataclass(frozen=True, **SLOTS)
class RestrictCharacters:
    alphabet: str

    def __get_pydantic_core_schema__(
        self, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        if not self.alphabet:
            raise ValueError("Alphabet may not be empty")

        schema = handler(source)  # get the CoreSchema from the type / inner constraints
        if schema["type"] != "str" and not (
            schema["type"] == "function-after" and schema["schema"]["type"] == "str"
        ):
            raise TypeError("RestrictCharacters can only be applied to strings")

        return no_info_after_validator_function(
            self.validate,
            schema,
        )

    def validate(self, value: str) -> str:
        if any(c not in self.alphabet for c in value):
            raise ValueError(f"{value!r} is not restricted to {self.alphabet!r}")
        return value
